# Demonstrator for Byte offset decompression in OpenCL

import numpy
import fabio
import pyopencl, pyopencl.array
import time
import os
# os.environ["PYOPENCL_CTX"] = "1:0"
os.environ["PYOPENCL_COMPILER_OUTPUT"] = "1"


def decomp_vec(raw_n):
    "principle of implementation in numpy"
    size = raw_n.size
    lel = numpy.ones(size, dtype="uint8")
    mask8 = raw_n == -128
    lel[mask8] = 3
    for i in numpy.where(mask8)[0]:
        if (raw_n[i + 1] == 0) and (raw_n[i + 2] == -128):
            lel[i] = 7
            print(i)
    lem = numpy.zeros_like(lel)
    return lel


def profile(evt, cmt=""):
    evt.wait()
    print("%s Exec time: %.3fms" % (cmt, 1e-6 * (evt.profile.end - evt.profile.start)))


ctx = pyopencl.create_some_context(interactive=True)


fname = "testimages/run2_1_00148.cbf"
cbf = fabio.cbfimage.CbfImage()
data = fabio.open(fname).data
raw = cbf.read(fname, only_raw=True)
properties = pyopencl.command_queue_properties.PROFILING_ENABLE
# properties = None
queue = pyopencl.CommandQueue(ctx, properties=properties)


raw_n = numpy.fromstring(raw, dtype="int8")
size = raw_n.size
raw_d = pyopencl.array.to_device(queue, raw_n)
int_d = pyopencl.array.empty(queue, (size,), dtype="int32")
data_d = pyopencl.array.empty(queue, (data.size,), dtype="int32")
tmp1_d = pyopencl.array.zeros_like(data_d)
tmp2_d = pyopencl.array.zeros_like(data_d)
tmp3_d = pyopencl.array.zeros_like(data_d)
lem_d = pyopencl.array.empty_like(data_d)
zero_d = pyopencl.array.zeros(queue, shape=1, dtype="int32")

src = open("sandbox/cbf.cl").read()
prg = pyopencl.Program(ctx, src).build()

for i in range(11):
    WG = 1 << i
    print("#" * 80)
    print("WG: %s" % WG)
    la = pyopencl.LocalMemory(4 * WG)
    lb = pyopencl.LocalMemory(4 * WG)
    lc = pyopencl.LocalMemory(4 * WG)
#     ld = pyopencl.LocalMemory(4)
    debug1_d = pyopencl.array.zeros(queue, shape=WG, dtype="int32")
    debug2_d = pyopencl.array.zeros(queue, shape=WG, dtype="int32")
    debug3_d = pyopencl.array.zeros(queue, shape=WG, dtype="int32")
    size = data.size

    wgsum_d = pyopencl.array.zeros(queue, shape=WG, dtype="int32")

    t0 = time.time()
    data_d.set(data.ravel())

    size = data.size
    WS = (size + WG - 1) & ~(WG - 1)
    chunk = ((size + WG - 1) // WG + WG - 1) // WG
    zero_d.fill(0)
    tmp2_d.fill(0)

    evt = prg.comp_byte_offset1(queue, (WG * WG,), (WG,),
                                data_d.data, tmp2_d.data, numpy.uint32(size), numpy.uint32(chunk), wgsum_d.data, zero_d.data,
                                la, lb, lc, debug1_d.data, debug2_d.data, debug3_d.data)
    profile(evt, "comp_byte_offset1")

    # Create dest buffers
    tmp_cumsum = wgsum_d.get()
    dest_size = tmp_cumsum[-1]

    print("Start process: %s" % debug3_d)
    print("End process: %s" % debug2_d)
    print("Total Size: %s" % (dest_size))
    print("After small cumsum=%s" % (tmp_cumsum))
    print("Counters= %s" % (debug1_d))
    target_d = pyopencl.array.zeros(queue, (dest_size,), dtype="int8")

    evt = prg.comp_byte_offset2(queue, (WG * WG,), (WG,),
                                data_d.data, tmp2_d.data, wgsum_d.data, target_d.data,
                                numpy.uint32(size), numpy.uint32(dest_size), numpy.uint32(chunk))
    profile(evt, "comp_byte_offset2")
    print("Total time : %.3fms" % (1000 * (time.time() - t0)))
    print("residual error: %s" % (numpy.where(raw_n - target_d.get())))



#!/usr/bin/env python3
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2015-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Run the tests of the project.

This script expects a suite function in <project_package>.test,
which returns a unittest.TestSuite.

Test coverage dependencies: coverage, lxml.
"""

__authors__ = ["Jérôme Kieffer", "Thomas Vincent"]
__date__ = "03/03/2023"
__license__ = "MIT"

import logging
import os
from argparse import ArgumentParser
import sys
import time
import unittest
import collections


class StreamHandlerUnittestReady(logging.StreamHandler):
    """The unittest class TestResult redefine sys.stdout/err to capture
    stdout/err from tests and to display them only when a test fail.
    This class allow to use unittest stdout-capture by using the last sys.stdout
    and not a cached one.
    """

    def emit(self, record):
        """
        :type record: logging.LogRecord
        """
        self.stream = sys.stderr
        super(StreamHandlerUnittestReady, self).emit(record)

    def flush(self):
        pass


def createBasicHandler():
    """Create the handler using the basic configuration"""
    hdlr = StreamHandlerUnittestReady()
    fs = logging.BASIC_FORMAT
    dfs = None
    fmt = logging.Formatter(fs, dfs)
    hdlr.setFormatter(fmt)
    return hdlr


# Use an handler compatible with unittests, else use_buffer is not working
logging.root.addHandler(createBasicHandler())

# Capture all default warnings
logging.captureWarnings(True)
import warnings
warnings.simplefilter('default')

logger = logging.getLogger("run_tests")
logger.setLevel(logging.WARNING)

logger.info("Python %s %s", sys.version, tuple.__itemsize__ * 8)
if sys.version_info.major < 3:
    logger.error("FabIO no more support Python2")

try:
    import resource
except ImportError:
    resource = None
    logger.warning("resource module missing")

try:
    import importlib
    importer = importlib.import_module
except ImportError:

    def importer(name):
        module = __import__(name)
        # returns the leaf module, instead of the root module
        subnames = name.split(".")
        subnames.pop(0)
        for subname in subnames:
            module = getattr(module, subname)
            return module

try:
    import numpy
except Exception as error:
    logger.warning("Numpy missing: %s", error)
else:
    logger.info("Numpy %s", numpy.version.version)

try:
    import h5py
except Exception as error:
    logger.warning("h5py missing: %s", error)
else:
    logger.info("h5py %s", h5py.version.version)

PROJECT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, PROJECT_DIR)
from bootstrap import get_project_name, build_project
PROJECT_NAME = get_project_name(PROJECT_DIR)
logger.info("Project name: %s", PROJECT_NAME)


class TextTestResultWithSkipList(unittest.TextTestResult):
    """Override default TextTestResult to display list of skipped tests at the
    end
    """

    def printErrors(self):
        unittest.TextTestResult.printErrors(self)
        # Print skipped tests at the end
        self.printGroupedList("SKIPPED", self.skipped)

    def printGroupedList(self, flavour, errors):
        grouped = collections.OrderedDict()

        for test, err in errors:
            if err in grouped:
                grouped[err] = grouped[err] + [test]
            else:
                grouped[err] = [test]

        for err, tests in grouped.items():
            self.stream.writeln(self.separator1)
            for test in tests:
                self.stream.writeln("%s: %s" % (flavour, self.getDescription(test)))
            self.stream.writeln(self.separator2)
            self.stream.writeln("%s" % err)


class ProfileTextTestResult(unittest.TextTestRunner.resultclass):

    def __init__(self, *arg, **kwarg):
        unittest.TextTestRunner.resultclass.__init__(self, *arg, **kwarg)
        self.logger = logging.getLogger("memProf")
        self.logger.setLevel(min(logging.INFO, logging.root.level))
        self.logger.handlers.append(logging.FileHandler("profile.log"))

    def startTest(self, test):
        unittest.TextTestRunner.resultclass.startTest(self, test)
        if resource:
            self.__mem_start = \
                resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        self.__time_start = time.time()

    def stopTest(self, test):
        unittest.TextTestRunner.resultclass.stopTest(self, test)
        # see issue 311. For other platform, get size of ru_maxrss in "man getrusage"
        if sys.platform == "darwin":
            ratio = 1e-6
        else:
            ratio = 1e-3
        if resource:
            memusage = (resource.getrusage(resource.RUSAGE_SELF).ru_maxrss -
                        self.__mem_start) * ratio
        else:
            memusage = 0
        self.logger.info("Time: %.3fs \t RAM: %.3f Mb\t%s",
                         time.time() - self.__time_start, memusage, test.id())


def report_rst(cov, package, version="0.0.0", base=""):
    """
    Generate a report of test coverage in RST (for Sphinx inclusion)

    :param cov: test coverage instance
    :param str package: Name of the package
    :param str base: base directory of modules to include in the report
    :return: RST string
    """
    import tempfile
    fd, fn = tempfile.mkstemp(suffix=".xml")
    os.close(fd)
    cov.xml_report(outfile=fn)

    try:
        from lxml import etree
    except ImportError:
        import xml.etree.ElementTree as etree

    xml = etree.parse(fn)
    classes = xml.findall(".//class")

    line0 = "Test coverage report for %s" % package
    res = [line0, "=" * len(line0), ""]
    res.append("Measured on *%s* version %s, %s" %
               (package, version, time.strftime("%d/%m/%Y")))
    res += ["",
            ".. csv-table:: Test suite coverage",
            '   :header: "Name", "Stmts", "Exec", "Cover"',
            '   :widths: 35, 8, 8, 8',
            '']
    tot_sum_lines = 0
    tot_sum_hits = 0

    for cl in classes:
        name = cl.get("name")
        fname = cl.get("filename")
        if not os.path.abspath(fname).startswith(base):
            continue
        lines = list(cl.find("lines"))
        hits = [int(i.get("hits")) for i in lines]

        sum_hits = sum(hits)
        sum_lines = len(lines)

        cover = 100.0 * sum_hits / sum_lines if sum_lines else 0

        if base:
            name = os.path.relpath(fname, base)

        res.append('   "%s", "%s", "%s", "%.1f %%"' %
                   (name, sum_lines, sum_hits, cover))
        tot_sum_lines += sum_lines
        tot_sum_hits += sum_hits
    res.append("")
    res.append('   "%s total", "%s", "%s", "%.1f %%"' %
               (package, tot_sum_lines, tot_sum_hits,
                100.0 * tot_sum_hits / tot_sum_lines if tot_sum_lines else 0))
    res.append("")
    return os.linesep.join(res)


def is_debug_python():
    """Returns true if the Python interpreter is in debug mode."""
    try:
        import sysconfig
    except ImportError:  # pragma nocover
        # Python < 2.7
        import distutils.sysconfig as sysconfig

    if sysconfig.get_config_var("Py_DEBUG"):
        return True

    return hasattr(sys, "gettotalrefcount")


parser = ArgumentParser(description='Run the tests.')

parser.add_argument("--installed",
                    action="store_true", dest="installed", default=False,
                    help=("Test the installed version instead of" +
                          "building from the source"))
parser.add_argument("-c", "--coverage", dest="coverage",
                    action="store_true", default=False,
                    help=("Report code coverage" +
                          "(requires 'coverage' and 'lxml' module)"))
parser.add_argument("-m", "--memprofile", dest="memprofile",
                    action="store_true", default=False,
                    help="Report memory profiling")
parser.add_argument("-v", "--verbose", default=0,
                    action="count", dest="verbose",
                    help="Increase verbosity. Option -v prints additional " +
                         "INFO messages. Use -vv for full verbosity, " +
                         "including debug messages and test help strings.")

default_test_name = "%s.test.suite" % PROJECT_NAME
parser.add_argument("test_name", nargs='*',
                    default=(default_test_name,),
                    help="Test names to run (Default: %s)" % default_test_name)
options = parser.parse_args()
sys.argv = [sys.argv[0]]

test_verbosity = 1
use_buffer = True
if options.verbose == 1:
    logging.root.setLevel(logging.INFO)
    logger.info("Set log level: INFO")
    test_verbosity = 2
    use_buffer = False
elif options.verbose > 1:
    logging.root.setLevel(logging.DEBUG)
    logger.info("Set log level: DEBUG")
    test_verbosity = 2
    use_buffer = False

if options.coverage:
    logger.info("Running test-coverage")
    import coverage
    omits = ["*test*", "*third_party*", "*/setup.py",
             # temporary test modules (silx.math.fit.test.test_fitmanager)
             "*customfun.py", ]
    try:
        cov = coverage.Coverage(omit=omits)
    except AttributeError:
        cov = coverage.coverage(omit=omits)
    cov.start()

# Prevent importing from source directory
if (os.path.dirname(os.path.abspath(__file__)) ==
        os.path.abspath(sys.path[0])):
    removed_from_sys_path = sys.path.pop(0)
    logger.info("Patched sys.path, removed: '%s'", removed_from_sys_path)

if options.installed:  # Use installed version
    for bad_path in (".", os.getcwd(), os.path.abspath(".")):
        if bad_path in sys.path:
            sys.path.remove(bad_path)
    try:
        module = importer(PROJECT_NAME)
    except Exception:
        logger.error("Cannot run tests on installed version: %s not installed or raising error.",
                     PROJECT_NAME)
        raise
else:  # Use built source
    build_dir = build_project(PROJECT_NAME, PROJECT_DIR)

    sys.path.insert(0, build_dir)
    logger.warning("Patched sys.path, added: '%s'", build_dir)
    module = importer(PROJECT_NAME)

PROJECT_VERSION = getattr(module, 'version', '')
PROJECT_PATH = module.__path__[0]

# Run the tests
runnerArgs = {}
runnerArgs["verbosity"] = test_verbosity
runnerArgs["buffer"] = use_buffer
if options.memprofile:
    runnerArgs["resultclass"] = ProfileTextTestResult
else:
    runnerArgs["resultclass"] = TextTestResultWithSkipList
runner = unittest.TextTestRunner(**runnerArgs)

logger.warning("Test %s %s from %s",
               PROJECT_NAME, PROJECT_VERSION, PROJECT_PATH)

test_module_name = PROJECT_NAME + '.test'
logger.info('Import %s', test_module_name)
test_module = importer(test_module_name)

test_suite = unittest.TestSuite()

if not options.test_name:
    # Do not use test loader to avoid cryptic exception
    # when an error occur during import
    project_test_suite = getattr(test_module, 'suite')
    test_suite.addTest(project_test_suite())
else:
    test_suite.addTest(
        unittest.defaultTestLoader.loadTestsFromNames(options.test_name))

# Display the result when using CTRL-C
unittest.installHandler()

result = runner.run(test_suite)

if result.wasSuccessful():
    exit_status = 0
else:
    exit_status = 1

if options.coverage:
    cov.stop()
    cov.save()
    with open("coverage.rst", "w") as fn:
        fn.write(report_rst(cov, PROJECT_NAME, PROJECT_VERSION, PROJECT_PATH))

sys.exit(exit_status)



# -*- coding: utf-8 -*-
#cython: embedsignature=True, language_level=3
## This is for optimisation
#cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False,
## This is for developping:
##cython: profile=True, warn.undeclared=True, warn.unused=True, warn.unused_result=False, warn.unused_arg=True
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) 2015 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
New Cython version of mar345_IO for preparing the migration to Python3

Compressor & decompressor for "pack" algorithm by JPA, binding to CCP4 libraries
those libraries are re-implemented in Cython.

Known bugs:
-----------

The precomp/postdec part need to be performed operation in int16 and exports
uint16. Some calculation are overflowing, this is needed to reproduce the
original implementation which is buggy

"""

__authors__ = ["Jerome Kieffer", "Gael Goret", "Thomas Vincent"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "2012-2020, European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "25/10/2023"

from libc.stdint cimport int8_t, uint8_t, int16_t, uint16_t, int32_t, uint32_t, int64_t, uint64_t 

import cython


import numpy
import os
import tempfile
import logging
logger = logging.getLogger(__name__)


ctypedef fused any_int_t:
    int8_t
    int16_t
    int32_t
    int64_t

# Few constants:
cdef:
    uint8_t *CCP4_PCK_BIT_COUNT = [0, 4, 5, 6, 7, 8, 16, 32]
    uint8_t *CCP4_BITSIZE = [0, 0, 0, 0, 1, 2, 3, 4, 5, 0, 0, 0, 0, 0, 0, 0,
                                 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7]
    unsigned int CCP4_PCK_BLOCK_HEADER_LENGTH = 6


cdef extern from "ccp4_pack.h":
    void* mar345_read_data_string(char *instream, int ocount, int dim1, int dim2) nogil
    void pack_wordimage_c(short int*, int , int , char*) nogil
    void* ccp4_unpack_string   (void *, void *, size_t, size_t, size_t) nogil
    void* ccp4_unpack_v2_string(void *, void *, size_t, size_t, size_t) nogil
cdef int PACK_SIZE_HIGH = 8


@cython.boundscheck(False)
def compress_pck(image not None, bint use_CCP4=False):
    """
    :param image: numpy array as input
    :param use_CCP4: use the former LGPL implementation provided by CCP4
    :return: binary stream
    """
    cdef:
        uint32_t  size, dim0, dim1, i, j
        int fd, ret
        char* name
        int16_t[::1] data
        int32_t[::1] raw
        bytes output

    assert image.ndim == 2, "Input image shape is 2D"
    size = image.size
    dim0 = image.shape[0]
    dim1 = image.shape[1]
    data = numpy.ascontiguousarray(image.ravel(), dtype=numpy.int16)
    if use_CCP4:
        (fd, fname) = tempfile.mkstemp()
        os.close(fd)
        fname = fname.encode("ASCII")
        name = <char*> fname
        with nogil:
            pack_wordimage_c(<short int *> &data[0], dim1, dim0, name)
        with open(name, "rb") as f:
            f.seek(0)
            output = f.read()
        os.unlink(fname)
    else:
        output = ("\nCCP4 packed image, X: %04d, Y: %04d\n" % (dim1, dim0)).encode("ASCII")
        raw = precomp(data, dim1)
        cont = pack_image(raw, False)
        output += cont.get().tobytes()
    return output


@cython.boundscheck(False)
@cython.cdivision(True)
def uncompress_pck(bytes raw not None, dim1=None, dim2=None, overflowPix=None, version=None, normal_start=None, swap_needed=None, bint use_CCP4=False):
    """
    Unpack a mar345 compressed image

    :param raw: input string (bytes in python3)
    :param dim1,dim2: optional parameters size
    :param overflowPix: optional parameters: number of overflowed pixels
    :param version: PCK version 1 or 2
    :param normal_start: position of the normal value section (can be auto-guessed)
    :param swap_needed: set to True when reading data from a foreign endianness (little on big or big on little)
    :return: ndarray of 2D with the right size
    """
    cdef:
        int cdimx, cdimy, chigh, cversion, records, normal_offset, lenkey, i, stop, idx, value
        uint32_t[:, ::1] data
        uint8_t[::1] instream
        int32_t[::1] unpacked
        int32_t[:, ::1] overflow_data  # handles overflows
        void* out
    end = None
    key1 = b"CCP4 packed image, X: "
    key2 = b"CCP4 packed image V2, X: "

    if (dim1 is None) or (dim2 is None) or \
       (version not in [1, 2]) or \
       (version is None) or \
       (normal_start is None):
        start = raw.find(key2)
        key = key2
        cversion = 2
        if start == -1:
            start = raw.find(key1)
            key = key1
            cversion = 1
        lenkey = len(key)
        start = raw.index(key) + lenkey
        sizes = raw[start:start + 13]
        cdimx = < int > int(sizes[:4])
        cdimy = < int > int(sizes[13-4:13])
        normal_offset = start + 13
    else:
        cdimx = < int > dim1
        cdimy = < int > dim2
        cversion = <int> version
        normal_offset = <int> normal_start
        if cversion == 1:
            lenkey = len(key1)
        else:
            lenkey = len(key2)
    if cversion not in [1, 2]:
        raise RuntimeError("Cannot determine the compression scheme for PCK compression (either version 1 or 2)")
    if (overflowPix is None) and (overflowPix is not False):
        end = raw.find("END OF HEADER")
        start = raw[:end].find("HIGH")
        hiLine = raw[start:end]
        hiLine = hiLine.split("\n")[0]
        word = hiLine.split()
        if len(word) > 1:
            chigh = int(word[1])
        else:
            logger.warning("Error while looking for overflowed pixels in line %s", hiLine.strip())
            chigh = 0
    else:
        chigh = < int > overflowPix

    instream = numpy.frombuffer(raw[normal_offset:].lstrip(), dtype=numpy.uint8).copy()

    if use_CCP4:
        data = numpy.empty((cdimy, cdimx), dtype=numpy.uint32)
        with nogil:
            ################################################################################
            #      rely to whichever version of ccp4_unpack is appropriate
            ################################################################################
            if cversion == 1:
                ccp4_unpack_string(&data[0, 0], &instream[0], cdimx, cdimy, 0)
            else:
                # cversion == 2:
                ccp4_unpack_v2_string(&data[0, 0], &instream[0], cdimx, cdimy, 0)
    else:
        # There is a bug in the mar345 implementation which performs arithmetics
        # of post-decompression in 16bits integers and overflows with large values
        unpacked = unpack_pck(instream, cdimx, cdimy).get1d()
        data = numpy.ascontiguousarray(postdec(unpacked, cdimx), numpy.uint32).reshape((cdimy, cdimx))

    if chigh > 0:
        ################################################################################
        # handle overflows: Each record is 8 overflow of 2x32bits integers
        ################################################################################
        records = (chigh + PACK_SIZE_HIGH - 1) // PACK_SIZE_HIGH
        stop = normal_offset - lenkey - 14
        odata = numpy.frombuffer(raw[stop - 64 * records: stop], dtype=numpy.int32).copy()
        if swap_needed:
            odata.byteswap(True)
        overflow_data = odata.reshape((-1, 2))
        for i in range(overflow_data.shape[0]):
            idx = overflow_data[i, 0] - 1     # indexes are even values (-1 because 1 based counting)
            value = overflow_data[i, 1]  # values are odd values
            if (idx >= 0) and (idx < cdimx * cdimy):
                data[idx // cdimx, idx % cdimx] = <uint32_t> value
    return numpy.asarray(data)


################################################################################
# Re-Implementation of the pck compression/decompression
################################################################################

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cpdef inline int32_t[::1] precomp(int16_t[::1] img, uint32_t width):
    """Pre-compression by subtracting the average value of the four neighbours

    Actually it looks a bit more complicated:

    * there comes the +2 from ?
    * the first element remains untouched
    * elements of the first line (+ first of second) use only former element


    JPA, the original author wrote:
    Compression is achieved by first calculating the differences between every
    pixel and the truncated value of four of its neighbours. For example:
    the difference for a pixel at img[x, y] is:

    comp[y, x] =  img[y, x] - (img[y-1, x-1] + img[y-1, x] + img[y-1, x+1] + img[y, x-1]) / 4

    This part implements overlows of int16 as the reference implementation is buggy
    """
    cdef:
        int size, i
        int32_t[::1] comp
        int16_t last, cur, im0, im1, im2
    size = img.size
    comp = numpy.zeros(size, dtype=numpy.int32)

    with nogil:
        # First pixel
        comp[0] = last = im0 = img[0]
        im1 = img[1]
        im2 = img[2]

        # First line (+ 1 pixel)
        for i in range(1, width + 1):
            cur = img[i]
            comp[i] = cur - last
            last = cur

        # Rest of the image
        for i in range(width + 1, size):
            cur = img[i]
            comp[i] = <int16_t> (cur - (last + im0 + im1 + im2 + 2) // 4)
            last = cur
            im0 = im1
            im1 = im2
            im2 = img[i - width + 2]

    return comp


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cpdef inline uint32_t[::1] postdec(int32_t[::1] comp, int width):
    """Post decompression by adding the average value of the four neighbours

    Actually it looks a bit more complicated:

    * there comes the +2 from ?
    * the first element remains untouched
    * elements of the first line (+ fist of second) use only former element

    JPA, the original author wrote:
    Compression is achieved by first calculating the differences between every
    pixel and the truncated value of four of its neighbours. For example:
    the difference for a pixel at img[x, y] is:

    comp[y, x] =  img[y, x] - (img[y-1, x-1] + img[y-1, x] + img[y-1, x+1] + img[y, x-1]) / 4

    This part implementes overlows of int16 as the reference implementation is bugged
    """
    cdef:
        int size, i
        uint32_t[::1] img
        int16_t last, cur, fl0, fl1, fl2
    size = comp.size

    img = numpy.zeros(size, dtype=numpy.uint32)

    with nogil:

        # First pixel
        last = comp[0]
        img[0] = last

        # First line (+ 1 pixel)
        for i in range(1, width + 1):
            img[i] = cur = comp[i] + last
            last = cur

        # Rest of the image: not parallel in this case
        fl0 = img[0]
        fl1 = img[1]
        fl2 = img[2]
        for i in range(width + 1, size):
            # overflow expected here.
            cur = comp[i] + (last + fl0 + fl1 + fl2 + 2) // 4
            # ensures the data is cropped at 16 bits!
            img[i] = <uint16_t> cur
            last = cur
            fl0 = fl1
            fl1 = fl2
            fl2 = img[i - width + 2]

    return img


################################################################################
# Re-Implementation of the pck compression stuff
################################################################################


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
cpdef inline int calc_nb_bits(any_int_t[::1] data, uint32_t start, uint32_t stop) nogil:
    """Calculate the number of bits needed to encode the data

    :param data: input data, probably slices of a larger array
    :param start: start position
    :param stop: stop position
    :return: the needed number of bits to store the values

    Comment from JPA:
    .................

    Returns the number of bits necessary to encode the longword-array 'chunk'
    of size 'n' The size in bits of one encoded element can be 0, 4, 5, 6, 7,
    8, 16 or 32.
     """
    cdef:
        uint32_t size, maxsize, i, abs_data
        any_int_t read_data

    size = stop - start
    maxsize = 0
    for i in range(start, stop):
        read_data = data[i]
        abs_data = - read_data if read_data < 0 else read_data
        if abs_data > maxsize:
            maxsize = abs_data
    if maxsize == 0:
        return 0
    elif maxsize < 8:
        return size * 4
    elif maxsize < 16:
        return size * 5
    elif maxsize < 32:
        return size * 6
    elif maxsize < 64:
        return size * 7
    elif maxsize < 128:
        return size * 8
    elif maxsize < 32768:
        return size * 16
    else:
        return size * 32


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.initializedcheck(False)
def pack_image(img, bint do_precomp=True):
    """Pack an image into a binary compressed block

    :param img: input image as numpy.int16
    :param do_precomp: perform the subtraction to the 4 neighbours's average. False is for testing the packing only
    :return: 1D array of numpy.int8

    JPA wrote:
    ..........
    Pack image 'img', containing 'x * y' WORD-sized pixels into byte-stream
    """
    cdef:
        uint32_t nrow, ncol, size, stream_size
        int16_t[::1] input_image
        int32_t[::1] raw
        PackContainer container
        uint32_t i, position
        uint32_t nb_val_packed
        uint32_t current_block_size, next_bock_size

    if do_precomp:
        assert len(img.shape) == 2
        nrow = img.shape[0]
        ncol = img.shape[1]
        input_image = numpy.ascontiguousarray(img, dtype=numpy.int16).ravel()
        # pre compression: subtract the average of the 4 neighbours
        raw = precomp(input_image, ncol)
        size = nrow * ncol
    else:
        raw = numpy.ascontiguousarray(img, dtype=numpy.int32).ravel()
        size = raw.size

    # allocation of the output buffer
    container = PackContainer(size)

    position = 0
    while position < size:
        nb_val_packed = 1
        current_block_size = calc_nb_bits(raw, position, position + nb_val_packed)
        while ((position + nb_val_packed) < size) and (nb_val_packed < 128):
            if (position + 2 * nb_val_packed) < size:
                next_bock_size = calc_nb_bits(raw, position + nb_val_packed, position + 2 * nb_val_packed)
            else:
                break
            if 2 * max(current_block_size, next_bock_size) < (current_block_size + next_bock_size + CCP4_PCK_BLOCK_HEADER_LENGTH):
                nb_val_packed *= 2
                current_block_size = 2 * max(current_block_size, next_bock_size)
            else:
                break
        container.append(raw, position, nb_val_packed, current_block_size)
        position += nb_val_packed

    return container


cdef class PackContainer:
    cdef:
        readonly uint32_t position, offset, allocated
        uint8_t[::1] data

    def __cinit__(self, uint32_t size=4096):
        """Constructor of the class

        :param size: start size of the array
        """
        self.position = 0
        self.offset = 0
        self.allocated = size
        self.data = numpy.zeros(self.allocated, dtype=numpy.uint8)

    def __dealloc__(self):
        self.data = None

    def get(self):
        """retrieve the populated array"""
        if self.offset:
            end = self.position + 1
        else:
            end = self.position
        return numpy.asarray(self.data[:end])

    cpdef append(self, int32_t[::1] data, uint32_t position, uint32_t nb_val, uint32_t block_size):
        """Append a block of data[position: position+nb_val] to the compressed
        stream. Only the most significant bits are takes.

        :param data: input uncompressed image as 1D array
        :param position: start position of reading of the image
        :param nb_val: number of value from data to pack in the block
        :param block_size: number of bits for the whole block

        The 6 bits header is managed here as well as the stream resizing.
        """
        cdef:
            uint32_t offset, index, i, bit_per_val, nb_bytes
            uint64_t tmp, tostore, mask
            int64_t topack

        bit_per_val = block_size // nb_val

        # realloc memory if needed
        nb_bytes = (CCP4_PCK_BLOCK_HEADER_LENGTH + block_size + 7) // 8
        if self.position + nb_bytes >= self.allocated:
            self.allocated *= 2
            new_stream = numpy.zeros(self.allocated, dtype=numpy.uint8)
            if self.offset:
                new_stream[:self.position + 1] = self.data[:self.position + 1]
            else:
                new_stream[:self.position] = self.data[:self.position]
            self.data = new_stream

        with nogil:
            if self.offset == 0:
                tmp = 0
            else:
                tmp = self.data[self.position]

            # append 6 bits of header
            tmp |= pack_nb_val(nb_val, bit_per_val) << self.offset
            self.offset += CCP4_PCK_BLOCK_HEADER_LENGTH
            self.data[self.position] = tmp & (255)
            if self.offset >= 8:
                self.position += 1
                self.offset -= 8
                self.data[self.position] = (tmp >> 8) & (255)

            # now pack every value in input stream"
            for i in range(nb_val):
                topack = data[position + i]

                mask = ((1 << (bit_per_val - 1)) - 1)
                tmp = (topack & mask)
                if topack < 0:
                    # handle the sign
                    tmp |= 1 << (bit_per_val - 1)

                # read last position
                if self.offset == 0:
                    tostore = 0
                else:
                    tostore = self.data[self.position]

                tostore |= tmp << self.offset
                self.offset += bit_per_val

                # Update the array
                self.data[self.position] = tostore & (255)
                while self.offset >= 8:
                    tostore = tostore >> 8
                    self.offset -= 8
                    self.position += 1
                    self.data[self.position] = tostore & (255)


cpdef inline uint8_t pack_nb_val(uint8_t nb_val, uint8_t value_size) nogil:
    """Calculate the header to be stored in 6 bits

    :param nb_val: number of values to be stored: must be a power of 2 <=128
    :param value_size: can be 0, 4, 5, 6, 7, 8, 16 or 32, the number of bits per value
    :return: the header as an unsigned char
    """
    cdef:
        uint32_t value, i

    value = 0
    for i in range(8):
        if (nb_val >> i) == 1:
            value |= i
            break
    value |= (CCP4_BITSIZE[value_size]) << (CCP4_PCK_BLOCK_HEADER_LENGTH >> 1)
    # should be 6/2 = 3
    return value


################################################################################
# Re-Implementation of the pck uncompression stuff
################################################################################
cpdef UnpackContainer unpack_pck(uint8_t[::1] stream, int ncol, int nrow):
    """Unpack the raw stream and return the image
    V1 only for now, V2 may be added later

    :param stream: raw input stream
    :param ncol: number of columns in the image (i.e width)
    :param nrow: number if rows in the image (i.e. height)
    :return: Container with decompressed image
    """
    cdef:
        uint32_t offset        # Number of bit to offset in the current byte
        uint32_t pos, end_pos  # current position and last position of block  in byte stream
        uint32_t size          # size of the input stream
        int32_t value, next    # integer values
        uint32_t nb_val_packed, nb_bit_per_val, nb_bit_in_block
        UnpackContainer cont       # Container with unpacked data

    cont = UnpackContainer(ncol, nrow)
    size = stream.size

    # Luckily we start at byte boundary
    offset = 0
    pos = 0

    while pos < (size) and cont.position < (cont.size):
        value = stream[pos]
        if offset > (8 - CCP4_PCK_BLOCK_HEADER_LENGTH):
            # wrap around
            pos += 1
            next = stream[pos]
            value |= next << 8
            value = value >> offset
            offset += CCP4_PCK_BLOCK_HEADER_LENGTH - 8
        elif offset == (8 - CCP4_PCK_BLOCK_HEADER_LENGTH):
            # Exactly on the boundary
            value = value >> offset
            pos += 1
            offset = 0
        else:
            # stay in same byte
            value = value >> offset
            offset += CCP4_PCK_BLOCK_HEADER_LENGTH

        # we use 7 as mask: decimal value of 111

        # move from offset, read 3 lsb, take the power of 2
        nb_val_packed = 1 << (value & 7)
        # read 3 next bits, search in LUT for the size of each element in block
        nb_bit_per_val = CCP4_PCK_BIT_COUNT[(value >> 3) & 7]

        if nb_bit_per_val == 0:
            cont.set_zero(nb_val_packed)
        else:
            nb_bit_in_block = nb_bit_per_val * nb_val_packed
            cont.unpack(stream, pos, offset, nb_val_packed, nb_bit_per_val)
            offset += nb_bit_in_block
            pos += offset // 8
            offset %= 8
    return cont


cdef class UnpackContainer:
    cdef:
        readonly uint32_t nrow, ncol, position, size
        readonly int32_t[::1] data
        # readonly list debug

    def __cinit__(self, int ncol, int nrow):
        self.nrow = nrow
        self.ncol = ncol
        self.size = nrow * ncol
        self.data = numpy.zeros(self.size, dtype=numpy.int32)
        self.position = 0

    def __dealloc__(self):
        self.data = None

    def get(self):
        """retrieve the populated array"""
        return numpy.asarray(self.data).reshape((self.nrow, self.ncol))

    cpdef int32_t[::1] get1d(self):
        """retrieve the populated array"""
        return self.data

    cpdef set_zero(self, int number):
        "set so many zeros"
        self.position += number

    cpdef unpack(self, uint8_t[::1] stream, uint32_t pos, uint32_t offset, uint32_t nb_value, uint32_t value_size):
        """unpack a block on data, all the same size

        :param stream: input stream, already sliced
        :param offset: number of bits of offset, at the begining of the stream
        :param nb_value: number of values to unpack
        :param value_size: number of bits of each value
        """
        cdef:
            uint32_t i           # simple counters
            int j
            uint32_t new_offset  # position after read
            int64_t cur          # value to be stored
            uint64_t tmp         # under contruction: needs to be unsigned
            int to_read          # number of bytes to read

        with nogil:
            cur = 0
            for i in range(nb_value):

                # read as many bytes as needed and unpack them to tmp variable

                tmp = stream[pos] >> offset

                new_offset = value_size + offset
                to_read = (new_offset + 7) // 8

                for j in range(1, to_read):
                    tmp |= (stream[pos + j]) << (8 * j - offset)

                # Remove the msb of tmp to keep only the interesting values
                cur = tmp & ((1 << (value_size)) - 1)

                # change sign if most significant bit is 1
                if cur >> (value_size - 1):
                    cur |= (-1) << (value_size - 1)

                # Update the storage
                self.data[self.position] = cur
                self.position += 1

                # Update the position in the array
                pos = pos + new_offset // 8
                offset = new_offset % 8



# coding: utf-8
#
#    Project: CCD mask image reader/writer. Format used in CrysalisPro software
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation files
#  (the "Software"), to deal in the Software without restriction,
#  including without limitation the rights to use, copy, modify, merge,
#  publish, distribute, sublicense, and/or sell copies of the Software,
#  and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

"""The CCD fileformat is used by the CrysalisPro software (provided for free by Rigaku)
The main developper of CrysalisPro is Mathias Meyer. Thanks to him for providing the structure of the file.
This file contains the code to generate a mask-file.
"""

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "2022 ESRF"
__date__ = "03/03/2023"

import logging
logger = logging.getLogger(__name__)
import time
import os
import numpy
from .fabioimage import FabioImage
from dataclasses import dataclass
from enum import Enum
import struct
# Some constants used in the file:
PAR_TEMPLATE = """#XCALIBUR SYSTEM
#XCALIBUR PARAMETER FILE - GENERATED BY THE COMMANDS WD PAR OR WD CAL
#THIS FILE SHOULD NOT BE EDITED BY HAND - USE THE PROGRAM KM4SETUP OR MODIFY PARS IN THE PROGRAM
#**********************************************************************************************
# D A T E  {date}
#**********************************************************************************************
#CELL INFORMATION
#      5.96002 (    0.00100 )     9.04296 (    0.00138 )    18.39626 (    0.00266 )
#     89.99769 (    0.01211 )    90.00650 (    0.01261 )    90.02161 (    0.01297 )
#  V =       991.49
#PROGRAM CRYSALIS 41.122a 64-bit (release 06-12-2021)
#MACHINE INFORMATION
#    - GONIOMETER KUMA/OXFORD INSTRUMENTS/OXFORD DIFFRACTION KM4
#   - ALPHA (DEG)   {alpha} BETA (DEG)    {beta}
#    - WAVELENGTH USER (ANG): A1    {wavelength} A2    {wavelength} B1    {wavelength}
#   - POLARISATION FACTOR    {polarization}#   - X-RAY BEAM ORIENTATION (DEG): X2   -0.01932 X3    0.00000
#   - BEAM PATH ABSORPTIONCOEFFICIENTS (MM-1): AIR    0.00013 BE    0.04736
#   - BE THICKNESS (MM):    0.50000
#   - SOFTWARE ZEROCORRECTION (DEG): OMEGA 0.04303 THETA -0.32229 KAPPA 0.00000 PHI 0.00000
#   - MACHINE OFFSETS (STEPS): OMEGA          0 THETA          0 KAPPA          0 PHI          0
#- CCD DETECTOR KM4CCD/SAPPHIRE
#   - CHIP: n/a n/a ({shape_x}x{shape_y}) n/a
#   - TAPER:  (,),
#            ZOOM 0.000000 PIXELSIZE IN 1X1 BINNING (MM)  0.06000
#   - GAIN 15.00 (DOUBLE-CORRELATED-DIVIDE-BY-2 GAIN)
#   - DETECTOR ROTATION (DEG): X1    0.00000 X2    0.00000 X3    0.00000
#   - DETECTOR DISTANCE (MM):   {distance}
#   - DETECTOR ZERO (PIX, 1X1 BINNING): X  512 Y  512
#   - DETECTOR BINNING (PIX): X:   1 Y:   1
#   - SETUP PATH : {path}
#   - GEOMETRIC CORRECTION FILE : NONE
#   - FLAT FIELD CORRECTION FILE: NONE
#   - GEOMETRIC CORRECTION STATUS: ALLOCATED:0, USED:0, GRIDX:0, GRIDY:0
#   - GEOMETRIC CORRECTION GRID LOADED: NONE
# CCD CHIPCONSTANTS systemgain darksignal readnoiserms overflowthreshold
CCD CHIPCONSTANTS 1.0 0.500 8.00 1000000
# CCD DARKACQUISITION numofdarkimages smpo_numofdarkimages type time
CCD DARKACQUISITION 4 2 DARK_TRIGGER_REF 60
# CCD DISPLAYOFFSET offset
CCD DISPLAYOFFSET       60
# CCD DISPLAYSHIFT shift
CCD DISPLAYSHIFT        -4
# CCD ISACARDSETTINGS iosettingno memsettingno interruptno
CCD ISACARDSETTINGS 3 2 5
# CCD PARAMETERS bx by x1 y1 xw yw dark flood spike readnode mpp gain
CCD PARAMETERS 1 1 0 0 {shape_x} {shape_y} DARKOFF FLOODOFF SPIKECORRELATION READNODEAD FIP60 GAINBOTH
# CCD VISUALIZATION lookuptablename colortablename
CCD VISUALIZATION LUT1 MORNINGGLORY
# CRYSTALLOGRAPHY ALPHAANGLEINDEG alpha alphapd delta deltapd
CRYSTALLOGRAPHY ALPHAANGLEINDEG   50.00000   50.0000   90.00000   90.00000
# CRYSTALLOGRAPHY BETAANGLEINDEG beta betapd gamma gammapd
CRYSTALLOGRAPHY BETAANGLEINDEG   0.00000   0.0000   90.00000   90.00000
# CRYSTALLOGRAPHY DATAREDUCTION C1 C2 LLthresh searchwinhalf LLprofpts LLstart LLback minwin ibackframes
CRYSTALLOGRAPHY DATAREDUCTION    2.50000    0.00000    6.00000     2    30     1     1     3     5
# CRYSTALLOGRAPHY DATRED3D #harvestref inxwin inywin
CRYSTALLOGRAPHY DATRED3D 1000 48 48
# CRYSTALLOGRAPHY EWALDEXPLORERCAMERA frx fry frz tox toy toz upx upy upz aph apv zo front back
CRYSTALLOGRAPHY EWALDEXPLORERCAMERA 1.0 0.0 0.0 -1.0 0.0 0.0 0.0 0.0 2.0 15.0 15.0 0.2 0 3
# CRYSTALLOGRAPHY EWALDEXPLORERSCREEN centerh centerv sizeh sizev
CRYSTALLOGRAPHY EWALDEXPLORERSCREEN 255 255 512 512
# CRYSTALLOGRAPHY NIGGLINUM ordinal niggli
CRYSTALLOGRAPHY NIGGLINUM 0 35
# CRYSTALLOGRAPHY NIGGLINUM ordinal niggli
CRYSTALLOGRAPHY NIGGLINUM 1 31
# CRYSTALLOGRAPHY NUMOFCELLS num
CRYSTALLOGRAPHY NUMOFCELLS        2
# CRYSTALLOGRAPHY NUMOFCELLSELECTED num
CRYSTALLOGRAPHY NUMOFCELLSELECTED        1
# CRYSTALLOGRAPHY ORIENTATIONMATRIX ub11 ub12 ub13 ... ub33 sub11 sub12 sub13 ... sub33
CRYSTALLOGRAPHY ORIENTATIONMATRIX -1.6894770E-002 -7.3228995E-002 -4.2883594E-002 7.9930593E-002 -2.0799582E-003 -3.9145127E-003 1.4021523E-002 -7.6538277E-002 4.1110883E-002 1.3594009E-005 1.9420319E-005 8.9086477E-006 1.2428387E-005 1.7755118E-005 8.1447731E-006 1.3554416E-005 1.9363756E-005 8.8827005E-006
# CRYSTALLOGRAPHY PEAKHUNTINGSCANWIDTH C1 C2
CRYSTALLOGRAPHY PEAKHUNTINGSCANWIDTH    1.00000    0.00000
# CRYSTALLOGRAPHY PEAKLOCATIONTABLEENTRIES numofpeaks
CRYSTALLOGRAPHY PEAKLOCATIONTABLEENTRIES    1000000
# CRYSTALLOGRAPHY PEAKTABLEENTRIES numofpeaks
CRYSTALLOGRAPHY PEAKTABLEENTRIES     1000000
# CRYSTALLOGRAPHY REDFORM r1 r2 r3 r4 r5 r6
CRYSTALLOGRAPHY REDFORM   8.90584865699E+001   1.51732728652E+002   2.94091358333E+002   4.22511681824E+001   1.50500907943E-001   8.68068266216E-003
# CRYSTALLOGRAPHY SIGMAREDFORM sr1 sr2 sr3 sr4 sr5 sr6
CRYSTALLOGRAPHY SIGMAREDFORM   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000
# CRYSTALLOGRAPHY SIGMAUB sub11 sub12 sub13 sub21 sub22 sub23 sub31 sub32 sub33
CRYSTALLOGRAPHY SIGMAUB   1.84797091592E-010   3.77148797034E-010   7.93640042416E-011   1.54464804286E-010   3.15244220669E-010   6.63373285638E-011   1.83722182099E-010   3.74955035116E-010   7.89023675307E-011
# CRYSTALLOGRAPHY SIGMAUM sum11 sum12 sum13 sum21 sum22 sum23 sum31 sum32 sum33
CRYSTALLOGRAPHY SIGMAUM   4.30291094497E-012   4.26333622558E-012   1.11470710805E-012   4.26333622558E-012   1.68813835518E-011   2.22021542351E-012   1.11470710805E-012   2.22021542351E-012   1.12128155741E-012
# CRYSTALLOGRAPHY UB ub11 ub12 ub13 ub21 ub22 ub23 ub31 ub32 ub33
CRYSTALLOGRAPHY UB  -1.68947699218E-002  -7.32289948389E-002  -4.28835937346E-002   7.99305932828E-002  -2.07995820653E-003  -3.91451274126E-003   1.40215232844E-002  -7.65382772048E-002   4.11108832953E-002
# CRYSTALLOGRAPHY UD ud11 ud12 ud13 ud21 ud22 ud23 ud31 ud32 ud33
CRYSTALLOGRAPHY UD   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000   0.00000000000E+000
# CRYSTALLOGRAPHY UM um11 um12 um13 um21 um22 um23 um31 um32 um33
CRYSTALLOGRAPHY UM   6.87093610846E-003  -2.24851001684E-006   9.88056331109E-004  -2.24851001684E-006   1.12249197887E-002   1.90830537823E-006   9.88056331109E-004   1.90830537823E-006   3.54443074691E-003
# CRYSTALLOGRAPHY UPDATECELLDATA
CRYSTALLOGRAPHY UPDATECELLDATA
# CRYSTALLOGRAPHY WAVELENGTH type
CRYSTALLOGRAPHY WAVELENGTH 0.373800 0.373800 0.373800
# DETECTOR CHARACTERISTICS offsetdeg deadtime filter
DETECTOR CHARACTERISTICS 17.0 0.00001 6.75
# EXPERIMENT COLLIMATOR len dia type
EXPERIMENT COLLIMATOR 140 0.5 PINHOLE
# EXPERIMENT COMPOUNDNAME compoundname
EXPERIMENT COMPOUNDNAME "compound name"
# EXPERIMENT GENERATOR kV mA type
EXPERIMENT GENERATOR     50.000     45.000 SEALEDTUBE
# EXPERIMENT TEMPERATURE tempinK
EXPERIMENT TEMPERATURE    293.000
# EXPERIMENT USER name
EXPERIMENT USER "Dr. User Name"
# FILE CHIP filename
FILE CHIP "{ccd_file}"
# FILE FLOOD filename
FILE FLOOD "NONE"
# FILE GEO filename
FILE GEO "NONE"
# FILE VERSION versionmajor versionminor versionstring
FILE VERSION 41 122 "41.122a 64-bit (release 06-12-2021)"
# GEOMETRYCORRECTION GRIDBASE e1xmm e1ymm e2xmm e2ymm offsetxmm offsetymm rotangledeg
GEOMETRYCORRECTION GRIDBASE 0.508 0.0 0.0 0.508 0.0 0.0 2.0
# GEOMETRYCORRECTION SEARCHWINDOW basewidthpix widthpix
GEOMETRYCORRECTION SEARCHWINDOW 12 12
# GONIOMETER ANGLE_LOLIM omlolimdeg thlolimdeg kalolimdeg phlolimdeg
GONIOMETER ANGLE_LOLIM  -300.00   -93.00  -179.00  -360.00
# GONIOMETER ANGLE_RANGEO omrangeodeg thrangeodeg karangeodeg phrangeodeg
GONIOMETER ANGLE_RANGEO 0 0 0 0
# GONIOMETER ANGLE_UPLIM omuplimdeg thuplimdeg kauplimdeg phuplimdeg
GONIOMETER ANGLE_UPLIM   300.00   120.00   179.00   360.00
# GONIOMETER COMPORT numofcomport
GONIOMETER COMPORT        2
# GONIOMETER COOLSWITCH coolswitch
GONIOMETER COOLSWITCH        1
# GONIOMETER DACDEF hv ll wi ga pt deadtime hvm llm wim gam ptm deadtimem fin fimonref fiblt isfilter filter1 filter2 filter3
GONIOMETER DACDEF   646.2558     0.8000     1.7000   140.0000     0.0000 1.2500000E-005      0.0000     0.0000     0.0000     0.0000     0.0000 0.00000000000E+000   80000       0       0 TRUE    7.90000     1.00000     1.00000
# GONIOMETER KM4VIDEO numoffaces
GONIOMETER KM4VIDEO       40
# GONIOMETER MICROSCOPE type
GONIOMETER MICROSCOPE SIDEWAYS
# GONIOMETER MOTORBACKDIRECTION omdir thdir kadir phdir
GONIOMETER MOTORBACKDIRECTION 3 3 1 0
# GONIOMETER MOTORBACKSPEED ommusperstep thmusperstep kamusperstep phmusperstep
GONIOMETER MOTORBACKSPEED 300 300 300 300
# GONIOMETER MOTORBACKSTEPS steps
GONIOMETER MOTORBACKSTEPS      500
# GONIOMETER MOTORDETECTORCOMM omsteps thsteps kasteps phsteps
GONIOMETER MOTORDETECTORCOMM 30 36 61 34
# GONIOMETER MOTORDETECTORLIMIT steps
GONIOMETER MOTORDETECTORLIMIT     6400
# GONIOMETER MOTORDETECTORSPEED ommusperstep thmusperstep kamusperstep phmusperstep
GONIOMETER MOTORDETECTORSPEED 350 300 300 300
# GONIOMETER MOTORZEROZONE omdeg thdeg kadeg phdeg ax4 ax5 ax6 ax7 ax8 ax9 omdegpd thdegpd kadegpd phdegpd ax4pd ax5pd ax6pd ax7pd ax8pd ax9pd
GONIOMETER MOTORZEROZONE  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000
# GONIOMETER PHICORR0_170 c0 c10 c20 c30 c40 c50 c60 c70 c80 c90 c100 c110 c120 c130 c140 c150 c160 c170
GONIOMETER PHICORR0_170 0.00 10.00 20.00 30.00 40.00 50.00 60.00 70.00 80.00 90.05 100.00 110.00 120.00 130.00 140.00 150.00 160.00 170.00
# GONIOMETER PHICORR180_360 c180 c190 c200 c210 c220 c230 c240 c250 c260 c270 c280 c290 c300 c310 c320 c330 c340 c350 c360
GONIOMETER PHICORR180_360 180.00 190.00 200.00 210.00 220.00 230.00 240.00 250.00 260.00 270.00 280.00 290.00 300.00 310.00 320.00 330.00 340.00 350.00 360.00
# GONIOMETER SIMULATOR onoff
GONIOMETER SIMULATOR off
# GONIOMETER TYPE specifier bsorient. xtobsmm bsdiamm CCDType isfilter filter1 filter2 filter3 remeasuretype isfrontend isgen.check  ispdav. isareadetav. Ispd bsang bsx bsy bsh
GONIOMETER TYPE KUMA_KM4 TOP    15.00000    1.50000 BC2_SAPPHIRE   FALSE     1.00000    1.00000    1.00000   TIME FALSE FALSE TRUE TRUE FALSE     0.00000 0.0 0.0 1.50
# LOG ENABLE onoff
LOG ENABLE on
# MAR SIGMACONTROL sc sigmasc readnoise darknoise E1 E2 rejectsc
MAR SIGMACONTROL 1.0 4.0 1.0 0 1 1 1
# MAR TEST s1 s2 s3 s4 s5 d1 d2 d3 d4 d5 l1 l2 l3 l4 l5
MAR TEST NONE NONE NONE NONE NONE 0.0 0.0 0.0 0.0 0.0 0 0 0 0 0
# RESERVE TEST1 dscsv0 dsctTx dsctf dscwOa dscwOb dscwC dmobf lmosm lmoss dmosn1 dmosn2 dmosVmin iswsmode swcephi  swceom swcegiveup swceint swcehalfcut swsmiwidth
RESERVE TEST1   0.200000   0.000000   0.000000   1.000000   0.500000   1.001000   0.500000     0     0   3.000000  25.000000   0.050000     0 2.000 0.500   150 0.250     8 0.000
# RESERVE TEST2 dmabn11 dmabn12 dmabn13 dmabn21 dmabn22 dmabn23 dmabn31 dmabn32 dmabn33 lmass1 lmass2 lmass3
RESERVE TEST2   0.000000   0.000000   0.000000   0.000000   0.000000   0.000000   0.000000   0.000000   0.000000     0     1     2
# RESERVE TEST3 ddaDh ddaDv ddlv ldlN dtrTmin dtrTmax ltrNshell ltrNstop lHmin lHmax lKmin lKmax lLmin lLmax
RESERVE TEST3   0.000000   0.000000   0.000000 0   2.000000  40.000000 0 0   0.000000   0.000000   0.000000   0.000000   0.000000   0.000000
# RESERVE TEST4 lrcn lrrn1 lrrn2 lrrn3 drrdO drrf
RESERVE TEST4 0 0 0 0   0.000000   5.000000
# RESERVE TEST5 darrh1 darrk1 darrl1 darrt1 darrfil1 darrh2 darrk2 darrl2 darrt2 darrfil2  darrh3 darrk3 darrl3 darrt3 darrfil3
RESERVE TEST5   0.000000   0.000000   0.000000   0.000000          0   0.000000   0.000000   0.000000   0.000000          0   0.000000   0.000000   0.000000   0.000000          0
# ROTATION ABSORPTIONCOEFFICIENTS wavelengthtype
ROTATION ABSORPTIONCOEFFICIENTS 0.000000 0.000000 0.000000
# ROTATION BEAM b2 b3 b2pd b3pd eqpd
ROTATION BEAM   0.000000    0.000000   0.00000   0.00000    0.00000
# ROTATION DETECTORORIENTATION d1 d2 d3 distanceinmm xorigininpix yorigininpix d1pd d2pd d3pd distanceinmmpd xorigininpixpd yorigininpixpd
ROTATION DETECTORORIENTATION    0.000000    0.000000    0.00000   {distance} {center_x} {center_y} -0.23494    0.11672    0.00000   50.65263  518.03346  502.88889
# ROTATION MONOCHROMATOR (dvalue|prepolfac) (orientation|type)
ROTATION MONOCHROMATOR    {polarization} SYNCHROTRON
# ROTATION PEAKINTEGRITYAREA enlargementfactor constantinpix userefMo userefCu refmod-closeMo,farMo,closeCu,farCu(d1,d2,t0,dd,x,y)
ROTATION PEAKINTEGRITYAREA 3.000 5.000 FALSE FALSE 0.745 0.401 0.766 60.000 505.164 493.255  0.745 0.401 0.766 100.000 505.164 493.255  0.745 0.401 0.766 60.000 505.164 493.255  0.745 0.401 0.766 100.000 505.164 493.255
# ROTATION SCANWIDTH c1_constant c2_thetadep ewaldconst ewaldlorentz
ROTATION SCANWIDTH {oscil} 0.000 1.000 0.000
# ROTATION ZEROCORRECTION omdeg thdeg kadeg phdeg ax4 ax5 ax6 ax7 ax8 ax9 omdegpd thdegpd kadegpd phdegpd ax4pd ax5pd ax6pd ax7pd ax8pd ax9pd
ROTATION ZEROCORRECTION  0.00000 0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000  0.00000
# ROTSIMULATOR ALPHAANGLEINDEG alpha
ROTSIMULATOR ALPHAANGLEINDEG       {alpha}
# ROTSIMULATOR BACKGROUNDNOISE typeofnoise
ROTSIMULATOR BACKGROUNDNOISE NONE
# ROTSIMULATOR BEAM b2 b3
ROTSIMULATOR BEAM 0.0 0.0
# ROTSIMULATOR BETAANGLEINDEG beta
ROTSIMULATOR BETAANGLEINDEG        {beta}
# ROTSIMULATOR CRYSTALLATTICE a b c alpha beta gamma
ROTSIMULATOR CRYSTALLATTICE 10.0 10.0 10.0 90.0 90.0 90.0
# ROTSIMULATOR CRYSTALORIENTATION r1 r2 r3
ROTSIMULATOR CRYSTALORIENTATION 0.0 0.0 0.0
# ROTSIMULATOR DETECTORORIENTATION d1 d2 d3 distanceinmm xorigininpix yorigininpix
ROTSIMULATOR DETECTORORIENTATION 0.0 0.0 0.0 {distance} {center_x} {center_y}
# ROTSIMULATOR DIFFRACTIONINTENSITY int dw
ROTSIMULATOR DIFFRACTIONINTENSITY 10000 0.0
# ROTSIMULATOR PEAKSHAPE xsizeinmm xsigmafactor ysizeinmm ysigmafactor
ROTSIMULATOR PEAKSHAPE 1.5 15.0 1.5 15.0
# ROTSIMULATOR SCANWIDTH c1_constant c2_thetadep omsigmafactor
ROTSIMULATOR SCANWIDTH 1.0 1.0 10.0
# ROTSIMULATOR SIMULATOR onoff
ROTSIMULATOR SIMULATOR off
# ROTSIMULATOR ZEROCORRECTION om th ka ph
ROTSIMULATOR ZEROCORRECTION 0.0 0.0 0.0 0.0
# SHAPE ABSORPTIONCOEFFICIENT coeffininvmm z
SHAPE ABSORPTIONCOEFFICIENT   0.000000   1.000000
# SHAPE CHEMICALFORMULA formula
SHAPE CHEMICALFORMULA "C10 H10 O5 N5"
# SHAPE FACE h k l dinmm
SHAPE FACE 0.0 0.0 0.0 0.0
# SHAPE VIDEOGRABBER grabber_type vfw_channel
SHAPE VIDEOGRABBER VFW 0
# SHAPE VIDEOSETUP cenx ceny scale
SHAPE VIDEOSETUP 384 288 0.00101890
# STOE TEST s1 s2 s3 s4 s5 d1 d2 d3 d4 d5 l1 l2 l3 l4 l5
STOE TEST NONE NONE NONE NONE NONE 0.0 0.0 0.0 0.0 0.0 0 0 0 0 0
#END SETUP XCALIBUR SYSTEM
"""


class CCD_FILEVERSION(Enum):
    VERS_1_1 = 0x00010001
    VERS_1_2 = 0x00010002
    VERS_1_3 = 0x00010003
    VERS_1_4 = 0x00010004
    VERS_1_5 = 0x00010005
    VERS_1_6 = 0x00010006
    VERS_1_7 = 0x00010007
    VERS_1_8 = 0x00010008
    VERS_1_9 = 0x00010009
    VERS_1_10 = 0x0001000a
    VERS_1_11 = 0x0001000b
    VERS_1_12 = 0x0001000c


CCD_FILEVERSION_VERS_HIGHEST = CCD_FILEVERSION.VERS_1_12


class CHIPCHARACTERISTICS_TREATMENT(Enum):
    IGNORE = 0
    REPLACE = 1
    AVERAGE = 2
    REPLACELEFT = 3
    REPLACERIGHT = 4
    AVERAGELEFTRIGHT = 5
    TREATMENT_REPLACETOP = 6
    REPLACEBOTTOM = 7
    AVERAGETOPBOTTOM = 8


class CHIPCHARACTERISTICS_POLYGONTYPE(Enum):
    TRIANGLE = 0
    RECTANGLE = 1
    POLYGON = 2
    MAXPOINTS = 6


class CHIPCHARACTERISTICS_SCINTILLATORID(Enum):
    NOTPRESENT = 0
    GREEN400 = 1
    LEXEL40 = 2
    LEXEL60 = 3
    LEXEL100 = 4
    RAREXMED = 5
    RAREXFINE = 6
    GREEN400_NEW = 7


CHIPCHARACTERISTICS_SCINTILLATORID_FIRST = CHIPCHARACTERISTICS_SCINTILLATORID.GREEN400
CHIPCHARACTERISTICS_SCINTILLATORID_LAST = CHIPCHARACTERISTICS_SCINTILLATORID.GREEN400_NEW


class SCAN_TYPE(Enum):
        Omega = 0
        Phi = 4


@dataclass
class ChipPoint:
    ix: int = 0
    iy: int = 0
    SIZE = 4

    @classmethod
    def loads(cls, buffer):
        assert len(buffer) >= cls.SIZE
        return cls(*struct.unpack("<HH", buffer[:cls.SIZE]))

    def dumps(self):
        return struct.pack("<HH", self.ix, self.iy)


@dataclass
class ChipBadPoint:
    spt: ChipPoint
    sptreplace1:ChipPoint
    sptreplace2: ChipPoint
    itreatment:int = 0
    SIZE = 14

    @classmethod
    def loads(cls, buffer):
        assert len(buffer) >= cls.SIZE
        return cls(ChipPoint.loads(buffer[:4]),
                   ChipPoint.loads(buffer[4:8]),
                   ChipPoint.loads(buffer[8:12]),
                   struct.unpack("<H", buffer[12:cls.SIZE]))

    def dumps(self):
        return  self.spt.dumps() + \
                self.sptreplace1.dumps() + \
                self.sptreplace2.dumps() + \
                struct.pack("<H", self.itreatment)


@dataclass
class ChipBadRow:
    """ROW STARTX ENDX Y"""
    sptstart: ChipPoint
    sptend: ChipPoint
    sptstartreplace:ChipPoint
    sptendreplace: ChipPoint
    itreatment:int
    SIZE = 18

    @classmethod
    def loads(cls, buffer):
        cls.SIZE = cls.SIZE
        assert len(buffer) >= cls.SIZE
        return cls(ChipPoint.loads(buffer[:4]),
                   ChipPoint.loads(buffer[4:8]),
                   ChipPoint.loads(buffer[8:12]),
                   ChipPoint.loads(buffer[12:16]),
                   struct.unpack("<H", buffer[16:cls.SIZE])[0])

    def dumps(self):
        return  self.sptstart.dumps() + \
                self.sptend.dumps() + \
                self.sptstartreplace.dumps() + \
                self.sptendreplace.dumps() + \
                struct.pack("<H", self.itreatment)


@dataclass
class ChipBadColumn:
    """COLUMN X STARTY ENDY"""
    sptstart: ChipPoint
    sptend: ChipPoint
    sptstartreplace:ChipPoint
    sptendreplace: ChipPoint
    itreatment:int
    ilowlimit: int
    ihighlimit: int
    SIZE = 22

    @classmethod
    def loads(cls, buffer):
        assert len(buffer) >= cls.SIZE
        return cls(ChipPoint.loads(buffer[:4]),
                   ChipPoint.loads(buffer[4:8]),
                   ChipPoint.loads(buffer[8:12]),
                   ChipPoint.loads(buffer[12:16]),
                   *struct.unpack("<HHH", buffer[16:cls.SIZE]))

    def dumps(self):
        return self.sptstart.dumps() + \
               self.sptend.dumps() + \
               self.sptstartreplace.dumps() + \
               self.sptendreplace.dumps() + \
               struct.pack("<HHH", self.itreatment, self.ilowlimit, self.ihighlimit)


@dataclass
class ChipBadPolygon:
    itype:int
    ipoints: int
    iax: list
    iay: list
    SIZE = 2 * (2 + 2 * CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value)

    @classmethod
    def loads(cls, buffer):
        assert len(buffer) >= cls.SIZE
        lst = struct.unpack("<HH" + "H"*2 * CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value,
                            buffer[:cls.SIZE])
        return cls(lst[0], lst[1],
                   lst[2:2 + CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value],
                   lst[2 + CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value:],)

    def dumps(self):
        for lst in (self.iax, self.iay):
            if len(lst) < CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value:
                lst += [0] * (CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value - len(lst))
        return struct.pack("<HH" + "H"*2 * CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value,
                           self.itype, self.ipoints,
                           *self.iax[:CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value],
                           *self.iay[:CHIPCHARACTERISTICS_POLYGONTYPE.MAXPOINTS.value])


@dataclass
class ChipMachineFunction:
    """
    xl is log10 of true Xray intensity
    y=A * exp (B * xl)
    """
    iismachinefunction: int = 0
    da_machinefct: float = 0.0
    db_machinefct: float = 0.0
    SIZE = 18

    @classmethod
    def loads(cls, buffer):
        assert len(buffer) >= cls.SIZE
        lst = struct.unpack("<Hdd",
                            buffer[:cls.SIZE])
        return cls(lst[0], lst[1], lst[2])

    def dumps(self):
        return struct.pack("<Hdd", self.iismachinefunction, self.da_machinefct, self.db_machinefct)


@dataclass
class Sweep:
    "Describes one scan or sweep, continuous rotation of one motor"
    iscanindex: int = 0
    iscantype: int = 0
    dwunknown1: int = 0
    domega: float = 0.0
    dtheta: float = 0.0
    dkappa: float = 0.0
    dphi: float = 0.0
    dstart: float = 0.0
    dend: float = 0.0
    dwidth: float = 0.0
    dunknown2: float = 0.0
    iunknown3: int = 0.0;
    iunknown4: int = 0.0;
    iunknown5: int = 0.0;
    iunknown6: int = 0.0;
    dexposure: float = 0.0
    SIZE = 88

    @classmethod
    def loads(cls, buffer):
        assert len(buffer) >= cls.SIZE
        lst = struct.unpack("<HHIddddddddHHHHd",
                            buffer[:cls.SIZE])
        return cls(*lst)

    def dumps(self):
        return struct.pack("<HHIddddddddHHHHd", self.iscanindex, self.iscantype,
                           self.dwunknown1, self.domega, self.dtheta, self.dkappa,
                           self.dphi, self.dstart, self.dend, self.dwidth,
                           self.dunknown2, self.iunknown3, self.iunknown4, self.iunknown5, self.iunknown6,
                           self.dexposure)


@dataclass
class RunDescription:
    """Names are using the hugarian notation: the first letter describes the type
    
    strings are 256bytes long
    floats are 64bits
    integers are 16bits, probably unsigned
    """
    cprefix: str = "n/a"
    cfolder: str = "n/a"
    inumofsweeps: int = 0
    dunknown: float = 0.0
    pssweep:list = tuple()
    SIZE = 528

    @classmethod
    def read(cls, filename):
        """Read the filename.run"""
        with open(filename, "rb") as f:
            bytestream = f.read()
        return cls.loads(bytestream)

    @classmethod
    def loads(cls, buffer):
        assert len(buffer) >= cls.SIZE
        prefix = buffer[:256].decode().strip("\x00")
        path = buffer[256:512].decode().strip("\x00")
        lst = struct.unpack("<Qd", buffer[512:cls.SIZE])
        self = cls(prefix, path, *lst, [])
        size = cls.SIZE
        for _ in range(self.inumofsweeps):
            self.pssweep.append(Sweep.loads(buffer[size:size + Sweep.SIZE]))
            size += Sweep.SIZE
        return self

    def save(self, filename):
        with open(filename, "wb")  as w:
            w.write(self.dumps())

    def dumps(self):

        # Some helper functions
        def record_str(key):
            value = self.__getattribute__(key)
            buffer[end:end + len(value)] = value.encode()
            return 256

        def record_struct(key, dtype):
            value = self.__getattribute__(key)
            size = struct.calcsize(dtype)
            if isinstance(value, (list, tuple)):
                buffer[end:end + size] = struct.pack(dtype, *value)
            else:
                buffer[end:end + size] = struct.pack(dtype, value)
            return size

        self.inumofsweeps = len(self.pssweep)

        size = self.SIZE + self.inumofsweeps * Sweep.SIZE
        buffer = bytearray(size)
        end = 0
        end += record_str("cprefix")
        end += record_str("cfolder")
        end += record_struct("inumofsweeps", "<Q")
        end += record_struct("dunknown", "<d")
        for sweep in self.pssweep:
            binary = sweep.dumps()
            length = len(binary)
            start = end
            end += length
            buffer[start:end] = binary
        return bytes(buffer[:end])


@dataclass
class CcdCharacteristiscs:
    """Names are using the hugarian notation: the first letter describes the type
    
    strings are 256bytes long
    floats are 64bits
    integers are 16bits, probably unsigned
    """
    dwversion: int = 0
    ddarkcurrentinADUpersec:float = 0.0
    dreadnoiseinADU: float = 0.0
    ccharacteristicsfil: str = "n/a"
    cccdproducer: str = "n/a"
    cccdchiptype: str = "n/a"
    cccdchipserial: str = "n/a"
    ctaperproducer: str = "n/a"
    ctapertype: str = "n/a"
    ctaperserial: str = "n/a"

    iisfip60origin: int = 0
    ifip60xorigin: int = 0
    ifip60yorigin: int = 0

    inumofcornermasks: int = 0
    iacornermaskx: list = tuple()
    iacornermasky:list = tuple()
    inumofglowingcornermasks: int = 0
    iaglowingcornermaskx: list = tuple()
    iaglowingcornermasky:list = tuple()

    ibadpolygons: int = 0
    pschipbadpolygon:list = tuple()

    ibadpoints: int = 0
    pschipbadpoint: list = tuple()

    ibadcolumns: int = 0
    pschipbadcolumn: list = tuple()
    ibadcolumns1x1: int = 0
    pschipbadcolumn1x1: list = tuple()
    ibadcolumns2x2: int = 0
    pschipbadcolumn2x2: list = tuple()
    ibadcolumns4x4: int = 0
    pschipbadcolumn4x4: list = tuple()

    ibadrows: int = 0
    pschipbadrow: list = tuple()

    iscintillatorid: int = 0
    # iisscintillatorid: int =0
    dgain_cu:float = 0.0
    dgain_mo:float = 0.0
    chipmachinefunction: ChipMachineFunction = None

    ibadrows1x1: int = 0
    pschipbadrow1x1: list = tuple()
    ibadrows2x2: int = 0
    pschipbadrow2x2: list = tuple()
    ibadrows4x4: int = 0
    pschipbadrow4x4: list = tuple()

    def _clip_string(self):
        """Clip all strings to 256 chars"""
        for key in ("ccharacteristicsfil", "cccdproducer", "cccdchiptype", "cccdchipserial", "ctaperproducer", "ctapertype", "ctaperserial"):
            value = self.__getattribute__(key)
            l = len(value)
            if l > 256:
                self.__setattr__(key, value[:256])

    @classmethod
    def read(cls, filename):
        """The the filename.ccd"""
        with open(filename, "rb") as f:
            bytestream = f.read()
        return cls.loads(bytestream)

    @classmethod
    def loads(cls, bytestream):
        ended = False
        length = len(bytestream)
        dico = {}
        if length > 1854:
            dico["dwversion"] = struct.unpack("I", bytestream[:4])[0]  # VERSION
            dico["ddarkcurrentinADUpersec"] = struct.unpack("d", bytestream[4:12])[0]  # DARK CURRENT IN ADU
            dico["dreadnoiseinADU"] = struct.unpack("d", bytestream[12:20])[0]  # READ NOSE IN ADU
            dico["ccharacteristicsfil"] = bytestream[20:276].decode().strip("\x00")  # CHARACTISTICS FILE NAME
            dico["cccdproducer"] = bytestream[276:532].decode().strip("\x00")  # PRODUCER
            dico["cccdchiptype"] = bytestream[532:788].decode().strip("\x00")  # CHIP TYPE
            dico["cccdchipserial"] = bytestream[788:1044].decode().strip("\x00")  # CHIP SERIAL
            dico["ctaperproducer"] = bytestream[1044:1300].decode().strip("\x00")  # TAPER PRODUCER
            dico["ctapertype"] = bytestream[1300:1556].decode().strip("\x00")  #  TAPER TYPE
            dico["ctaperserial"] = bytestream[1556:1812].decode().strip("\x00")  #  TAPER SERIAL
            dico["iisfip60origin"], dico["ifip60xorigin"], dico["ifip60yorigin"] = struct.unpack("HHH", bytestream[1812:1818])  # FIP60ORIGIN

            dico["inumofcornermasks"] = struct.unpack("H", bytestream[1818:1820])  # CORNER MASKS
            dico["iacornermaskx"] = struct.unpack("HHHH", bytestream[1820:1828])
            dico["iacornermasky"] = struct.unpack("HHHH", bytestream[1828:1836])

            dico["inumofglowingcornermasks"] = struct.unpack("H", bytestream[1836:1838])  # GLOWINGCORNER MASKS
            dico["iaglowingcornermaskx"] = struct.unpack("HHHH", bytestream[1838:1846])
            dico["iaglowingcornermasky"] = struct.unpack("HHHH", bytestream[1846:1854])
        else:
            ended = True
        # NUMBER OF BAD POLYGONS
        start = 1856
        if not ended and start <= length:
            try:
                dico["ibadpolygons"] = struct.unpack("H", bytestream[1854:1856])[0]
                polygons = dico["pschipbadpolygon"] = []
                for _ in range(dico["ibadpolygons"]):
                    polygons.append(ChipBadPolygon.loads(bytestream[start:]))
                    start += ChipBadPolygon.SIZE
            except AssertionError:
                ended = True
        # NUMBER OF BAD POINTS
        if not ended and start + 2 <= length:
            try:
                dico["ibadpoints"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                points = dico["pschipbadpoint"] = []
                for _ in range(dico["ibadpoints"]):  # LOOP OVER BAD POINTS
                    points.append(ChipBadPoint.loads(bytestream[start:]))
                    start += ChipBadPoint.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD COLS
        if not ended and start + 2 <= length:
            try:
                dico["ibadcolumns"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadcolumn"] = []
                for _ in range(dico["ibadcolumns"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadColumn.loads(bytestream[start:]))
                    start += ChipBadColumn.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD COLS 1X1
        if not ended and start + 2 <= length:
            try:
                dico["ibadcolumns1x1"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadcolumn1x1"] = []
                for _ in range(dico["ibadcolumns1x1"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadColumn.loads(bytestream[start:]))
                    start += ChipBadColumn.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD COLS 2X2
        if not ended and start + 2 <= length:
            try:
                dico["ibadcolumns2x2"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadcolumn2x2"] = []
                for _ in range(dico["ibadcolumns2x2"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadColumn.loads(bytestream[start:]))
                    start += ChipBadColumn.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD COLS 4X4
        if not ended and start + 2 <= length:
            try:
                dico["ibadcolumns4x4"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadcolumn4x4"] = []
                for _ in range(dico["ibadcolumns4x4"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadColumn.loads(bytestream[start:]))
                    start += ChipBadColumn.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD ROWS
        if not ended and start + 2 <= length:
            try:
                dico["ibadrows"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadrow"] = []
                for _ in range(dico["ibadrows"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadRow.loads(bytestream[start:]))
                    start += ChipBadRow.SIZE
            except AssertionError:
                ended = True

        if not ended and start + 18 <= length:
            try:
                dico["iscintillatorid"] = struct.unpack("H", bytestream[start:start + 2])[0]  # SCINTILLATOR DESCRIPTION
                dico["dgain_mo"] = struct.unpack("d", bytestream[start + 2:start + 10])[0]  # GAINMO
                dico["dgain_cu"] = struct.unpack("d", bytestream[start + 10:start + 18])[0]  # GAINCU
                start += 18
                dico["chipmachinefunction"] = ChipMachineFunction.loads(bytestream[start:])  # IISMACHINEFUNCTION
                start += ChipMachineFunction.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD ROWS 1X1
        if not ended and start + 2 <= length:
            try:
                dico["ibadrows1x1"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadrow1x1"] = []
                for _ in range(dico["ibadcolumns1x1"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadRow.loads(bytestream[start:]))
                    start += ChipBadRow.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD ROWS 2X2
        if not ended and start + 2 <= length:
            try:
                dico["ibadrows2x2"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadrow2x2"] = []
                for _ in range(dico["ibadrows2x2"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadRow.loads(bytestream[start:]))
                    start += ChipBadRow.SIZE
            except AssertionError:
                ended = True

        # NUMBER OF BAD ROWS 4X4
        if not ended and start + 2 <= length:
            try:
                dico["ibadrows4x4"] = struct.unpack("H", bytestream[start:start + 2])[0]
                start += 2
                columns = dico["pschipbadrow4x4"] = []
                for _ in range(dico["ibadrows4x4"]):  # LOOP OVER BAD COLS
                    columns.append(ChipBadRow.loads(bytestream[start:]))
                    start += ChipBadRow.SIZE
            except AssertionError:
                ended = True

        self = cls(**dico)
        return self

    def save(self, filename):
        with open(filename, "wb")  as w:
            w.write(self.dumps())

    def dumps(self):
        """Dump the content of the struct as a bytestream."""
        buffer = bytearray(4096)
        end = 0
        # prepare the structure
        if not self.dwversion:
            self.dwversion = CCD_FILEVERSION_VERS_HIGHEST
        self._clip_string()
        self.ibadpolygons = len(self.pschipbadpolygon)
        self.ibadpoints = len(self.pschipbadpoint)
        for empty_4_tuple in ("iacornermaskx", "iacornermasky",
                              "iaglowingcornermaskx", "iaglowingcornermasky"):
            value = self.__getattribute__(empty_4_tuple)
            if len(value) == 0:
                self.__setattr__(empty_4_tuple, [0, 0, 0, 0])
        if self.chipmachinefunction is None:
            self.chipmachinefunction = ChipMachineFunction()

        # Some helper functions
        def record_str(key):
            value = self.__getattribute__(key)
            buffer[end:end + len(value)] = value.encode()
            return 256

        def record_struct(key, dtype):
            value = self.__getattribute__(key)
            size = struct.calcsize(dtype)
            if isinstance(value, (list, tuple)):
                buffer[end:end + size] = struct.pack(dtype, *value)
            else:
                buffer[end:end + size] = struct.pack(dtype, value)
            return size

        def record_object(key):
            value = self.__getattribute__(key)
            tmp = value.dumps()
            ltmp = len(tmp)
            buffer[end:end + ltmp] = tmp
            return ltmp

        def record_variable(key, subkey, dtype="H"):
            size = struct.calcsize(dtype)
            values = self.__getattribute__(subkey)
            nitems = len(values)
            self.__setattr__(key, nitems)
            buffer[end:end + size] = struct.pack(dtype, nitems)
            for i in values:
                tmp = i.dumps()
                ltmp = len(tmp)
                buffer[end + size:end + size + ltmp] = tmp
                size += ltmp
            return size

        end += record_struct("dwversion", "I")  # VERSION
        end += record_struct("ddarkcurrentinADUpersec", "d")  # DARK CURRENT IN ADU
        end += record_struct("dreadnoiseinADU", "d")  # READ NOSE IN ADU
        end += record_str("ccharacteristicsfil")
        end += record_str("cccdproducer")
        end += record_str("cccdchiptype")
        end += record_str("cccdchipserial")
        end += record_str("ctaperproducer")
        end += record_str("ctapertype")
        end += record_str("ctaperserial")
        end += record_struct("iisfip60origin", "H")
        end += record_struct("ifip60xorigin", "H")
        end += record_struct("ifip60yorigin", "H")
        end += record_struct("inumofcornermasks", "H")
        end += record_struct("iacornermaskx", "HHHH")
        end += record_struct("iacornermasky", "HHHH")
        end += record_struct("inumofglowingcornermasks", "H")
        end += record_struct("iaglowingcornermaskx", "HHHH")
        end += record_struct("iaglowingcornermasky", "HHHH")
        end += record_variable("ibadpolygons", "pschipbadpolygon")
        end += record_variable("ibadpoints", "pschipbadpoint")
        end += record_variable("ibadcolumns", "pschipbadcolumn")
        end += record_variable("ibadcolumns1x1", "pschipbadcolumn1x1")
        end += record_variable("ibadcolumns2x2", "pschipbadcolumn2x2")
        end += record_variable("ibadcolumns4x4", "pschipbadcolumn4x4")
        end += record_variable("ibadrows", "pschipbadrow")
        end += record_struct("iscintillatorid", "H")
        end += record_struct("dgain_mo", "d")
        end += record_struct("dgain_cu", "d")
        end += record_object("chipmachinefunction")
        end += record_variable("ibadrows1x1", "pschipbadrow1x1")
        end += record_variable("ibadrows2x2", "pschipbadrow2x2")
        end += record_variable("ibadrows4x4", "pschipbadrow4x4")
        return bytes(buffer[:end])

    def build_mask(self, shape):
        """Rebuild a mask from all pixels/rectangles masked"""
        mask = numpy.zeros(shape, dtype=numpy.int8)
        for poly in self.pschipbadpolygon:
            if poly.itype == CHIPCHARACTERISTICS_POLYGONTYPE.RECTANGLE.value:
                mask[poly.iay[0]:1 + poly.iay[1], poly.iax[0]:1 + poly.iax[1]] = 1
        for pt in self.pschipbadpoint:
            mask[pt.spt.iy, pt.spt.ix] = 1
        return mask


class XcaliburImage(FabioImage):
    """FabIO image class for CrysalisPro mask image
    """

    DESCRIPTION = "Xcalibur binary struct of masked pixels"

    DEFAULT_EXTENSIONS = []

    def __init__(self, *arg, **kwargs):
        """
        """
        FabioImage.__init__(self, *arg, **kwargs)

    def _readheader(self, infile):
        """
        Read and decode the header of an image:

        :param infile: Opened python file (can be stringIO or bzipped file)
        """
        # list of header key to keep the order (when writing)
        self.header = self.check_header()

    def read(self, fname, frame=None):
        """
        Try to read image

        :param fname: name of the file
        :param frame: number of the frame
        """

        self.resetvals()
        with self._open(fname) as infile:
            self._readheader(infile)
            # read the image data and declare it

        shape = (50, 60)
        self.data = numpy.zeros(shape, dtype=self.uint16)
        # Nota: dim1, dim2, bytecode and bpp are properties defined by the dataset
        return self

    def decompose(self, full=False):
        """Decompose a mask defined as a 2D binary image in 
        
        * vertical+horizontal gaps
        * rectangles
        * bad pixels 
        
        :param: full: deal only with gaps (False) or perform the complete analysis (True) 
        :return: CcdCharacteristiscs struct.
        """
        ccd = CcdCharacteristiscs(CCD_FILEVERSION_VERS_HIGHEST.value,
                                  pschipbadpolygon=[],
                                  pschipbadpoint=[])
        mask = numpy.array(self.data, dtype=bool)
        shape = mask.shape

        row_gaps = self._search_gap(mask, dim=1)
        col_gaps = self._search_gap(mask, dim=0)

        ccd.ibadpolygons = len(row_gaps) + len(col_gaps)
        ccd.pschipbadpolygon = []
        for gap in row_gaps:
            polygon = ChipBadPolygon(CHIPCHARACTERISTICS_POLYGONTYPE.RECTANGLE.value, 4,
                                     [0, shape[1] - 1], [gap[0], gap[1] - 1])
            ccd.pschipbadpolygon.append(polygon)
        for gap in col_gaps:
            polygon = ChipBadPolygon(CHIPCHARACTERISTICS_POLYGONTYPE.RECTANGLE.value, 4,
                                     [gap[0], gap[1] - 1], [0, shape[0] - 1])
            ccd.pschipbadpolygon.append(polygon)

        try:
            import pyFAI.ext.dynamic_rectangle
        except ImportError:
            logger.warning("PyFAI not available: only a coarse description of the mask is provided")
            full = False
        if not full:
            return ccd
        # Decompose detector into a set of modules, then extract patches of mask for each of them:
        c = 0
        for cg in col_gaps + [(self.shape[1], self.shape[1])]:
            r = 0
            for rg in row_gaps + [(self.shape[0], self.shape[0])]:
                mm = mask[r:rg[0], c:cg[0]]
                if mm.size:
                    rectangles = pyFAI.ext.dynamic_rectangle.decompose_mask(mm)
                    for rec in rectangles:
                        if rec.area == 1:
                            point = ChipPoint(c + rec.col, r + rec.row)
                            bad_point = ChipBadPoint(point, point, point, CHIPCHARACTERISTICS_TREATMENT.IGNORE.value)
                            ccd.ibadpoints += 1
                            ccd.pschipbadpoint.append(bad_point)
                        else:
                            ccd.ibadpolygons += 1
                            polygon = ChipBadPolygon(CHIPCHARACTERISTICS_POLYGONTYPE.RECTANGLE.value, 4,
                                                     [c + rec.col, c + rec.col + rec.width - 1],
                                                     [r + rec.row, r + rec.row + rec.height - 1])
                            ccd.pschipbadpolygon.append(polygon)
                r = rg[1]
            c = cg[1]
        return ccd

    @staticmethod
    def _search_gap(mask, dim=0):

        shape = mask.shape
        m0 = numpy.sum(mask, axis=dim, dtype="int") == shape[dim]
        if m0.any():
            m0 = numpy.asarray(m0, "int8")
            d0 = m0[1:] - m0[:-1]
            starts = numpy.where(d0 == 1)[0]
            stops = numpy.where(d0 == -1)[0]
            if  (len(starts) == 0):
                starts = numpy.array([-1])
            if  (len(stops) == 0):
                stops = numpy.array([len(m0) - 1])
            if (stops[0] < starts[0]):
                starts = numpy.concatenate(([-1], starts))
            if (stops[-1] < starts[-1]):
                stops = numpy.concatenate((stops, [len(m0) - 1]))
            r0 = [ (start + 1, stop + 1) for start, stop  in zip(starts, stops)]
        else:
            r0 = []
        return r0

    def save_par(self, path, prefix, **kwargs):
        """Generate a *.par" file which contains the parameters of the scan 
        """

        dico = {"alpha": kwargs.get("alpha", 50),
                "beta": kwargs.get("beta", 0),
                "wavelength": kwargs.get("wavelength", 1),
                "polarization": kwargs.get("polarization", 0.98),
                "shape_x": self.shape[1],
                "shape_y": self.shape[0],
                "distance": kwargs.get("distance", 100),  # mm
                "path": path,
                "ccd_file": prefix + ".ccd",
                "oscil": kwargs.get("oscil", 1),
                "center_x":kwargs.get("center_x", self.shape[1] / 2),
                "center_y":kwargs.get("center_y", self.shape[1] / 2),
                "date": time.ctime()
                }
        string = PAR_TEMPLATE.format(**dico)
        blines = [b"\xa7" + i[1:].rstrip().encode() if i.startswith("#") else i.rstrip().encode() for i in string.split("\n")]
        with open(os.path.join(path, prefix + ".par"), "wb") as w:
            w.write(b"\r\n".join(blines))


# This is for compatibility with old code:
xcaliburimage = XcaliburImage



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#########################################
# Command line converter a bunch of files from CBF to eiger
# European Synchrotron Radiation Facility
#
#########################################
"""Converter a bunch of files from any format to an eiger-data
"""

__author__ = "Jérôme Kieffer"
__copyright__ = "2016 ESRF"
__date__ = "06/04/2020"
__licence__ = "MIT"

import logging
logging.basicConfig()

import os

import numpy
import fabio
from fabio import nexus

import argparse
from threading import Thread, Event

from queue import Queue

logger = logging.getLogger("to_eiger")


class Reader(Thread):
    """Reader with input and output queue 
    """

    def __init__(self, queue_in, queue_out, quit_event):
        """Constructor of the class
        
        :param queue_in: input queue with (index, filename to read) as input
        :param queue_out: output queue with (index, FabioImage) as output
        :param quit_event: the event which tells the thread to end 
        """
        Thread.__init__(self)
        self._queue_in = queue_in
        self._queue_out = queue_out
        self._quit_event = quit_event

    def run(self):
        while not self._quit_event.is_set():
            plop = self._queue_in.get()
            if plop is None:
                break
            idx, fname = plop
            try:
                fimg = fabio.open(fname)
            except Exception as err:
                logger.error(err)
                fimg = None
            self._queue_out.put((idx, fimg))
            self._queue_in.task_done()

    @classmethod
    def build_pool(cls, args, size=1):
        """Create a pool of worker of a given size. 
        
        :param worker: class of the worker (deriving  from Thread) 
        :param args: arguments to be passed to each of the worker
        :param size: size of the pool
        :return: a list of worker 
        """
        workers = []
        for _ in range(size):
            w = cls(*args)
            w.start()
            workers.append(w)
        return workers


def save_eiger(input_files, output_file, filter_=None, nbthreads=None):
    """Save a bunch of files in Eiger-like format
    
    :param input_files: list of input files
    :param output_file: name of the HDF5 file
    :param filter_: Type of compression filter: "gzip", "lz4" or "bitshuffle"
    :param nbthreads: number of parallel reader threads  
    """
    assert len(input_files), "Input file list is not empty"
    first_image = input_files[0]
    fimg = fabio.open(first_image)
    shape = fimg.data.shape
    stack_shape = (len(input_files),) + shape
    first_frame_timestamp = os.stat(first_image).st_ctime
    kwfilter = {}
    if filter_ == "gzip":
        kwfilter = {"compression": "gzip", "shuffle": True}
    elif filter_ == "lz4":
        kwfilter = {"compression": 32004, "shuffle": True}
    elif filter_ == "bitshuffle":
        kwfilter = {"compression": 32008, "compression_opts": (0, 2)}  # enforce lz4 compression

    if nbthreads:
        queue_in = Queue()
        queue_out = Queue()
        quit_event = Event()
        pool = Reader.build_pool((queue_in, queue_out, quit_event), nbthreads)
        for idx, fname in enumerate(input_files[1:]):
            queue_in.put((idx, fname))
    with nexus.Nexus(output_file) as nxs:
        entry = nxs.new_entry(entry='entry', program_name='fabio',
                              title='converted from single-frame files',
                              force_time=first_frame_timestamp,
                              force_name=True)
        data = nxs.new_class(grp=entry, name="data", class_type="NXdata")
        try:
            ds = data.require_dataset(name="data", shape=stack_shape,
                                      dtype=fimg.data.dtype,
                                      chunks=(1,) + shape,
                                      **kwfilter)
        except Exception as error:
            logger.error("Error in creating dataset, disabling compression:%s", error)
            ds = data.require_dataset(name="data", shape=stack_shape,
                                      dtype=fimg.data.dtype,
                                      chunks=(1,) + shape)

        ds[0] = fimg.data
        data["sources"] = [numpy.string_(i) for i in input_files]
        if nbthreads:
            for _ in range(len(input_files) - 1):
                idx, fimg = queue_out.get()
                if fimg.data is None:
                    logger.error("Failed reading file: %s", input_files[idx + 1])
                    continue
                ds[idx + 1] = fimg.data
                queue_out.task_done()

            queue_in.join()
            queue_out.join()
        else:  # don't use the pool of readers
            for idx, fname in enumerate(input_files[1:]):
                ds[idx + 1] = fabio.open(fname).data

    if nbthreads:  # clean up
        quit_event.set()
        for _ in pool:
            queue_in.put(None)


if __name__ == "__main__":
    epilog = "plop"
    parser = argparse.ArgumentParser(prog="cbf2eiger",
                                     description=__doc__,
                                     epilog=epilog)
    parser.add_argument("IMAGE", nargs="*",
                        help="Input file images")
    parser.add_argument("-V", "--version", action='version', version=fabio.version,
                        help="output version and exit")
    parser.add_argument("-v", "--verbose", action='store_true', dest="verbose", default=False,
                        help="show information for each conversions")
    parser.add_argument("--debug", action='store_true', dest="debug", default=False,
                        help="show debug information")

    group = parser.add_argument_group("main arguments")
    group.add_argument("-l", "--list", action="store_true", dest="list", default=None,
                       help="show the list of available formats and exit")
    group.add_argument("-o", "--output", dest='output', type=str,
                       help="output file or directory")
    group.add_argument("-f", "--filter", dest='filter', type=str, default=None,
                       help="Compression filter, may be lz4, bitshuffle or gzip")
    group.add_argument("-n", "--nbthreads", dest='nbthreads', type=int, default=None,
                       help="Numbre of reader threads in parallel")

    opts = parser.parse_args()
    input_files = [os.path.abspath(i) for i in opts.IMAGE if os.path.exists(i)]
    input_files.sort()
    save_eiger(input_files, opts.output, filter_=opts.filter, nbthreads=opts.nbthreads)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
2011: Jerome Kieffer for ESRF.

Unit tests for CBF images based on references images taken from:
http://pilatus.web.psi.ch/DATA/DATASETS/insulin_0.2/

19/01/2015
"""

import unittest
import os
import time
import logging

logger = logging.getLogger(__name__)
import numpy
import fabio
from fabio.cbfimage import CbfImage, CIF
from fabio.compression import decByteOffset_numpy, decByteOffset_cython
from ..utilstest import UtilsTest
from ..testutils import LoggingValidator


class TestCbfReader(unittest.TestCase):
    """ test cbf image reader """

    def setUp(self):
        """Download images"""
        self.edf_filename = "run2_1_00148.edf.bz2"
        self.edf_filename = UtilsTest.getimage(self.edf_filename)[:-4]
        self.cbf_filename = "run2_1_00148.cbf.bz2"
        self.cbf_filename = UtilsTest.getimage(self.cbf_filename)[:-4]

    def test_read(self):
        """ check whole reader"""
        times = []
        times.append(time.time())
        cbf = fabio.open(self.cbf_filename)
        times.append(time.time())
        edf = fabio.open(self.edf_filename)
        times.append(time.time())

        self.assertAlmostEqual(0, abs(cbf.data - edf.data).max())
        logger.info("Reading CBF took %.3fs whereas the same EDF took %.3fs" % (times[1] - times[0], times[2] - times[1]))

    def test_write(self):
        "Rest writing with self consistency at the fabio level"
        name = os.path.basename(self.cbf_filename)
        obj = CbfImage()
        obj.read(self.cbf_filename)
        obj.write(os.path.join(UtilsTest.tempdir, name))
        other = CbfImage()
        other.read(os.path.join(UtilsTest.tempdir, name))
        self.assertEqual(abs(obj.data - other.data).max(), 0, "data are the same")
        for key in obj.header:
            if key in["filename", "X-Binary-Size-Padding"]:
                continue
            self.assertTrue(key in other.header, "Key %s is in header" % key)
            self.assertEqual(obj.header[key], other.header[key], "value are the same for key %s" % key)
        # By destroying the object, one actually closes the file, which is needed under windows.
        del obj
        del other
        if os.path.exists(os.path.join(UtilsTest.tempdir, name)):
            os.unlink(os.path.join(UtilsTest.tempdir, name))

    def test_byte_offset(self):
        """ check byte offset algorithm"""
        cbf = fabio.open(self.cbf_filename)
        starter = b"\x0c\x1a\x04\xd5"
        cbs = cbf.cbs
        startPos = cbs.find(starter) + 4
        data = cbs[startPos: startPos + int(cbf.header["X-Binary-Size"])]
        startTime = time.time()
        size = cbf.shape[0] * cbf.shape[1]
        numpyRes = decByteOffset_numpy(data, size=size)
        tNumpy = time.time() - startTime
        logger.info("Timing for Numpy method : %.3fs" % tNumpy)

        startTime = time.time()
        cythonRes = decByteOffset_cython(stream=data, size=size)
        tCython = time.time() - startTime
        delta = abs(numpyRes - cythonRes).max()
        self.assertAlmostEqual(0, delta)
        logger.info("Timing for Cython method : %.3fs, max delta= %s" % (tCython, delta))

    def test_consitency_manual(self):
        """
        Test if an image can be read and saved and the results are "similar"
        """
        name = os.path.basename(self.cbf_filename)
        obj = fabio.open(self.cbf_filename)
        new = CbfImage(data=obj.data, header=obj.header)
        new.write(os.path.join(UtilsTest.tempdir, name))
        other = fabio.open(os.path.join(UtilsTest.tempdir, name))
        self.assertEqual(abs(obj.data - other.data).max(), 0, "data are the same")
        for key in obj.header:
            if key in["filename", "X-Binary-Size-Padding"]:
                continue
            self.assertTrue(key in other.header, "Key %s is in header" % key)
            self.assertEqual(obj.header[key], other.header[key], "value are the same for key %s [%s|%s]" % (key, obj.header[key], other.header[key]))

    def test_consitency_convert(self):
        """
        Test if an image can be read and saved and the results are "similar"
        """
        name = os.path.basename(self.cbf_filename)
        obj = fabio.open(self.cbf_filename)
        new = obj.convert("cbf")
        new.write(os.path.join(UtilsTest.tempdir, name))
        other = fabio.open(os.path.join(UtilsTest.tempdir, name))
        self.assertEqual(abs(obj.data - other.data).max(), 0, "data are the same")
        for key in obj.header:
            if key in["filename", "X-Binary-Size-Padding"]:
                continue
            self.assertTrue(key in other.header, "Key %s is in header" % key)
            self.assertEqual(obj.header[key], other.header[key], "value are the same for key %s [%s|%s]" % (key, obj.header[key], other.header[key]))

    def test_unicode(self):
        """
        Test if an image can be read and saved to an unicode named
        """
        name = u"%s" % os.path.basename(self.cbf_filename)
        obj = fabio.open(self.cbf_filename)
        obj.write(os.path.join(UtilsTest.tempdir, name))
        other = fabio.open(os.path.join(UtilsTest.tempdir, name))
        self.assertEqual(abs(obj.data - other.data).max(), 0, "data are the same")
        for key in obj.header:
            if key in["filename", "X-Binary-Size-Padding"]:
                continue
            self.assertTrue(key in other.header, "Key %s is in header" % key)
            self.assertEqual(obj.header[key], other.header[key], "value are the same for key %s [%s|%s]" % (key, obj.header[key], other.header[key]))

    def test_bug_388(self):
        data = numpy.random.randint(0, 100, size=(27, 31)).astype(numpy.int32)
        im = CbfImage(data=data, header={})
        filename = os.path.join(UtilsTest.tempdir, "bad_name.edf")
        im.write(filename)
        res = fabio.open(filename)
        self.assertEqual(abs(data - res.data).max(), 0, "Data are the same despite the wrong name")

    def test_bug_408(self):
        "Bug when reading files generated by XDS"
        xds_filename = UtilsTest.getimage("XDS.cbf.bz2")[:-4]
        _ = fabio.open(xds_filename)

    def test_bug_528(self):
        "Bug when writing/reading small files"
        filename = os.path.join(UtilsTest.tempdir, "bug528.cbf")
        data = numpy.array([[1, 2, 3]])
        CbfImage(header={}, data=data).write(filename)
        with LoggingValidator(fabio.cbfimage.logger, error=0, warning=0):
            fimg = fabio.open(filename)
        self.assertEqual(abs(data - fimg.data).max(), 0, "data match")

    def test_bug_535(self):
        "Bug when reading files with long path in header"
        cbf_filename = UtilsTest.getimage("ref-test_5_0001.cbf")
        fimg = fabio.open(cbf_filename)
        for key in ["_array_data.header_contents", "_array_data.header_convention"]:
            self.assertIn(key, fimg.header, f"{key} in header")

    def test_cif(self):
        LaB6 = """
data_global
_amcsd_formula_title 'B6La'
loop_
_publ_author_name
'Eliseev A'
'Efremmov V'
'Kuzmicheva G'
'Konovalova E'
'Lazorenko V'
'Paderno Y'
'Khlyustova S'
_journal_name_full 'Kristallografiya'
_journal_volume 31 
_journal_year 1986
_journal_page_first 803
_journal_page_last 805
_publ_section_title
;
 X-ray structural investigation of single crystals of lanthanum, cerium,
 and samarium hexaborides
 _cod_database_code 1000057
;
_database_code_amcsd 0014189
_chemical_formula_sum 'La B6'
_chemical_formula_sum ''
_cell_length_a 4.1570
_cell_length_b 4.1570
_cell_length_c 4.1570
_cell_angle_alpha 90
_cell_angle_beta 90
_cell_angle_gamma 90
_cell_volume 71.836
_exptl_crystal_density_diffrn      4.710
_symmetry_space_group_name_H-M 'P m 3 m'
loop_
_space_group_symop_operation_xyz
  'x,y,z'
  'z,-x,y'
  '-y,z,-x'
  'x,-y,z'
  '-z,x,-y'
  'y,-z,x'
  '-x,y,-z'
  'x,-z,-y'
  '-z,y,x'
  'y,-x,-z'
  '-x,z,y'
  'z,-y,-x'
  '-y,x,z'
  'x,z,y'
  '-z,-y,-x'
  'y,x,z'
  '-x,-z,-y'
  'z,y,x'
  '-y,-x,-z'
  'z,x,-y'
  '-y,-z,x'
  'x,y,-z'
  '-z,-x,y'
  'y,z,-x'
  '-x,-y,z'
  '-z,x,y'
  'y,-z,-x'
  '-x,y,z'
  'z,-x,-y'
  '-y,z,x'
  'x,-y,-z'
  '-x,z,-y'
  'z,-y,x'
  '-y,x,-z'
  'x,-z,y'
  '-z,y,-x'
  'y,-x,z'
  '-x,-z,y'
  'z,y,-x'
  '-y,-x,z'
  'x,z,-y'
  '-z,-y,x'
  'y,x,-z'
  '-z,-x,-y'
  'y,z,x'
  '-x,-y,-z'
  'z,x,y'
  '-y,-z,-x'
loop_
_atom_site_label
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
La   0.00000   0.00000   0.00000
B   0.19750   0.50000   0.50000
"""
        filename = os.path.join(UtilsTest.tempdir, "LaB6.cif")
        filename2 = os.path.join(UtilsTest.tempdir, "LaB6_2.cif")
        # print(filename, filename2)
        with open(filename, "w") as w:
            w.write(LaB6)
        lab6 = CIF(filename)
        # print(lab6)
        lab6.saveCIF(filename2)
        lab62 = CIF(filename2)
        self.assertEqual(len(lab6), len(lab62), "size matches")


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestCbfReader))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
Unit tests for xcalibur struct used by crysalis to represent the mask

"""
__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "2022 ESRF"
__date__ = "23/02/2023"

import unittest
import os
import time
import logging

logger = logging.getLogger(__name__)
import numpy
import fabio
from fabio.xcaliburimage import CcdCharacteristiscs, XcaliburImage
from ..utilstest import UtilsTest


class TestCcdCharacteristiscs(unittest.TestCase):
    """ test CcdCharacteristiscs struct reader """
    @classmethod
    def setUpClass(cls)->None:
        cls.ccdfiles = [UtilsTest.getimage(i) for i in ("scan0001.ccd", "scan0005.ccd", "scan0050.ccd", "scan0059.ccd", "scan0066.ccd")]
    @classmethod
    def tearDownClass(cls)->None:
        cls.ccfiles = None 
    def test_parse(self):
        for afile in self.ccdfiles:
            ref = CcdCharacteristiscs.read(afile)
            obt = CcdCharacteristiscs.loads(ref.dumps())
            for key in CcdCharacteristiscs.__dataclass_fields__:
                for what in (ref, obt):
                    if what.__getattribute__(key) == tuple():
                        what.__setattr__(key, [])
            self.assertEqual(ref, obt, f"{afile} matches ")
    
class testXcalibureImage(unittest.TestCase):
    @classmethod
    def setUpClass(cls)->None:
        cls.filename = UtilsTest.getimage("Pilatus1M.cbf")
    @classmethod
    def tearDownClass(cls)->None:
        cls.filename = None 
    def test_decomposition(self):
        ref = (fabio.open(self.filename).data<0).astype("int8")
        xcal = XcaliburImage(data=ref)
        try:
            import pyFAI.ext.dynamic_rectangle
        except ImportError:
            logger.warning("PyFAI not available: only a coarse description of the mask is provided")
            precise = False
        else:
            precise = True

        ccd = xcal.decompose(full=True)
        obt = ccd.build_mask(ref.shape)
        if precise:
            self.assertTrue(numpy.allclose(ref, obt), "mask is the same")        

    
def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestCcdCharacteristiscs))
    testsuite.addTest(loadTests(testXcalibureImage))    
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())
      
        


# -*- coding: utf-8 -*-
#cython: embedsignature=True, language_level=3
## This is for optimisation
#cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False,
## This is for developping:
##cython: profile=True, warn.undeclared=True, warn.unused=True, warn.unused_result=False, warn.unused_arg=True
#
#    Project: Fable Input/Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) 2020-2020 European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Compression and decompression extension for Esperanto format
"""
__author__ = "Jérôme Kieffer"
__date__ = "13/11/2020"
__contact__ = "Jerome.kieffer@esrf.fr"
__license__ = "MIT"

from io import BytesIO
from struct import pack
import numpy
from libc.stdint cimport int32_t, int64_t, uint64_t, uint32_t, uint8_t, uint16_t, int16_t
ctypedef fused any_int:
    int32_t
    int64_t

cpdef int32_t fieldsize(any_int nbvalue):
    "Direct translation of Fortran"
    return _fieldsize(nbvalue)

cdef uint8_t _fieldsize(any_int nbvalue) nogil:
    "Direct translation of Fortran"
    cdef uint8_t getfieldsize
    if(nbvalue < -63):
        getfieldsize = 8
    elif(nbvalue < -31):
        getfieldsize = 7
    elif(nbvalue < -15):
        getfieldsize = 6
    elif(nbvalue < -7):
        getfieldsize = 5
    elif(nbvalue < -3):
        getfieldsize = 4
    elif(nbvalue < -1):
        getfieldsize = 3
    elif(nbvalue < 0):
        getfieldsize = 2
    elif(nbvalue < 2):
        getfieldsize = 1
    elif(nbvalue < 3):
        getfieldsize = 2
    elif(nbvalue < 5):
        getfieldsize = 3
    elif(nbvalue < 9):
        getfieldsize = 4
    elif(nbvalue < 17):
        getfieldsize = 5
    elif(nbvalue < 33):
        getfieldsize = 6
    elif(nbvalue < 65):
        getfieldsize = 7
    else:
        getfieldsize = 8
    return getfieldsize


cpdef int32_t get_fieldsize(any_int[::1] array):
    """determine the fieldsize to store the given values
    
    :param array numpy.array
    :returns int
    """
    cdef:
        any_int maxi, mini, value
        int32_t size, idx, 
    maxi = mini = 0
    size = array.shape[0]
    for idx in range(size):
        value = array[idx]
        maxi = max(maxi, value)
        mini = min(mini, value)
    
    return max(fieldsize(maxi), fieldsize(mini))

cdef uint8_t _get_fieldsize(int32_t[::1] array) nogil:
    """determine the fieldsize to store the given values
    
    :param array numpy.array
    :returns int
    """
    cdef:
        int32_t maxi, mini, value
        int32_t size, idx, 
    maxi = mini = 0
    size = array.shape[0]
    for idx in range(size):
        value = array[idx]
        maxi = max(maxi, value)
        mini = min(mini, value)
    
    return max(_fieldsize(maxi), _fieldsize(mini))


cpdef write_escaped(int32_t value, buffer):
    """write an value to the buffer and escape when overflowing one byte
    :param value: int
    :param buffer: io.BytesIO
    """
    if -127 <= value < 127:
        buffer.write(pack("B", value + 127))
    elif -32767 < value < 32767:
        buffer.write(b'\xfe' + pack("<h", value))
    else:
        buffer.write(b'\xff' + pack("<i", value))

cdef uint16_t _write_escaped(int32_t value, uint8_t[::1] buffer, uint16_t position) nogil:
    """write an value to the buffer at the given position and escape when overflowing one byte
    :param value: int
    :param buffer: compressed line
    :param position: where to start writing in the buffer
    :return: now position in the vector 
    """
    cdef:
        int16_t value_16
        int32_t value_32
        int64_t one=1
        
    if -127 <= value < 127:
        buffer[position] = <uint8_t> (value + 127)
        return position+1
    elif -32767 < value < 32767:
        buffer[position] = 254
        value_16 = value 
        buffer[position+1] = <uint8_t>(value_16 & ((one<<8)-1))
        buffer[position+2] = <uint8_t>((value_16 & ((one<<16)-1))>>8)
        return position+3 
    else:
        buffer[position] = 255
        buffer[position+1] = <uint8_t>((value & ((one<<8)-1)))
        buffer[position+2] = <uint8_t>((value & ((one<<16)-1))>>8) 
        buffer[position+3] = <uint8_t>((value & ((one<<24)-1))>>16)
        buffer[position+4] = <uint8_t>((value & ((one<<32)-1))>>24)
        return position+5 


cpdef bytes compress_field(int32_t[::1] ifield, int32_t fieldsize, overflow_table):
    """compress a field with given size
    :param ifield: numpy.ndarray
    :param fieldsize: int
    :param overflow_table: io.BytesIO
    :returns: bytes
    """
    cdef:
        uint64_t compressed_field, i, val, conv_
        int32_t elem

    if fieldsize == 8:
        # we have to deal offsets but not bitshifts
        tmp = bytearray(8)
        for i in range(8):
            elem = ifield[i]
            if -127 <= elem < 127:
                tmp[i] = elem + 127
            elif -32767 <= elem < 32767:
                tmp[i] = 254
                overflow_table.write(pack("<h", elem))
            else:
                tmp[i] = 255
                overflow_table.write(pack("<i", elem))
        return bytes(tmp)
    elif fieldsize > 0:
        # we have to deal with bit-shifts but not offsets
        conv_ = (1<<(fieldsize - 1)) - 1 
        compressed_field = 0
        for i in range(8):
            val = ifield[i] + conv_
            compressed_field |= val << (i * fieldsize)
        try:
            res = pack("<Q", compressed_field)
        except:
            print("Exception in struct.pack: %s %s %s %s", fieldsize, type(fieldsize), ifield, compressed_field)
            raise
        return res[:fieldsize]
    else:
        raise AssertionError("fieldsize is between 0 and 8")
    
cdef uint16_t _compress_field(int32_t[::1] ifield, int32_t fieldsize, uint8_t[::1] buffer, uint16_t position, uint16_t overflow_position) nogil:
    """compress a field with given size
    :param ifield: numpy.ndarray with 8 data to be compressed
    :param fieldsize: int size of the field, number of bits per item
    :param buffer: the output buffer with compressed data
    :param position: Where to write the bitfield in the buffer
    :param overflow_position: Where to write the overflow table in the buffer
    :returns: position of th endd of the overflow values 
    """
    cdef:
        uint64_t compressed_field, i, val, conv_, one
        int32_t value
        int16_t value_16
    if fieldsize == 8:
        # we have to deal offsets but not bitshifts
        one = 1
        for i in range(8):
            value = ifield[i]
            if -127 <= value < 127:
                buffer[position+i] = value + 127
            elif -32767 <= value < 32767:
                buffer[position+i] = 254
                value_16 = value
                buffer[overflow_position] = <uint8_t>(value_16 & ((one<<8)-1))
                buffer[overflow_position+1] = <uint8_t>((value_16 & ((one<<16)-1))>>8)
                overflow_position += 2
            else:
                buffer[position+i] =  255
                buffer[overflow_position] = <uint8_t>((value & ((one<<8)-1)))
                buffer[overflow_position+1] = <uint8_t>((value & ((one<<16)-1))>>8) 
                buffer[overflow_position+2] = <uint8_t>((value & ((one<<24)-1))>>16)
                buffer[overflow_position+3] = <uint8_t>((value & ((one<<32)-1))>>24)
                overflow_position += 4
        return overflow_position
    elif fieldsize > 0:
        # we have to deal with bit-shifts but not offsets
        conv_ = (1<<(fieldsize - 1)) - 1 
        compressed_field = 0
        for i in range(8):
            val = ifield[i] + conv_
            compressed_field |= val << (i * fieldsize)
        for i in range(<uint64_t>fieldsize):
            buffer[position+i] =  <uint8_t>((compressed_field>>(i*8)) & 255)
        return overflow_position
    else:
        raise AssertionError("fieldsize is between 0 and 8")


def compress_row(int32_t[::1] data, buffer):
    """compress a single row
    :arg data numpy.array
    :arg buffer io.BytesIO
    """
    cdef:
        int32_t size, first_pixel, n_fields, n_restpx, len_a, len_b, cur, prev, i
        int32_t[::1] fielda, fieldb, pixel_diff
    
    size = data.shape[0]
    first_pixel = data[0]
    
    #data[1:] - data[:size-1]
    pixel_diff = numpy.empty(size-1, dtype=numpy.int32)
    prev = first_pixel
    for i in range(1, size):
        cur = data[i]
        pixel_diff[i-1] = cur - prev
        prev = cur

    write_escaped(first_pixel, buffer)

    n_fields = (size-1) // 16
    n_restpx = (size-1) % 16

    for i in range(n_fields):
        fielda = pixel_diff[:8]
        fieldb = pixel_diff[8:16]

        len_a = get_fieldsize(fielda)
        len_b = get_fieldsize(fieldb)
        len_byte = (len_b << 4) | len_a
        buffer.write(pack("B", len_byte))

        of_buff = BytesIO()
        compressed_fielda = compress_field(fielda, len_a, of_buff)
        compressed_fieldb = compress_field(fieldb, len_b, of_buff)

        buffer.write(compressed_fielda + compressed_fieldb + of_buff.getvalue())

        pixel_diff = pixel_diff[16:]

    for i in range(n_restpx):
        write_escaped(pixel_diff[i], buffer)
        
cdef uint16_t _compress_row(int32_t [::1] data, uint8_t[::1] buffer, int32_t[::1] pixel_diff) nogil:
    """compress a single row
    :param data: numpy.array one line of the input frame
    :param buffer: buffer where to store compressed data
    :return: size of the compressed line
    """
    cdef:
        int32_t size, first_pixel, n_fields, n_restpx, len_a, len_b, cur, prev, i
        int32_t[::1] fielda, fieldb
        uint16_t position, overflow_position
    
    size = data.shape[0]
    first_pixel = data[0]
    

    #data[1:] - data[:size-1]
    #pixel_diff = numpy.empty(size-1, dtype=numpy.int32)
    prev = first_pixel
    for i in range(1, size):
        cur = data[i]
        pixel_diff[i-1] = cur - prev
        prev = cur

    position = _write_escaped(first_pixel, buffer, 0)

    n_fields = (size-1) // 16
    n_restpx = (size-1) % 16

    for i in range(n_fields):
        fielda = pixel_diff[:8]
        fieldb = pixel_diff[8:16]

        len_a = _get_fieldsize(fielda)
        len_b = _get_fieldsize(fieldb)
        buffer[position] = (len_b << 4) | len_a
        position += 1
        
        overflow_position = _compress_field(fielda, len_a, buffer, position, position+len_a+len_b)
        position = _compress_field(fieldb, len_b, buffer, position+len_a, overflow_position)

        pixel_diff = pixel_diff[16:]

    for i in range(n_restpx):
        position = _write_escaped(pixel_diff[i], buffer, position)

    return position

def compress(int32_t [:,::1] frame):
    """compress a frame using the agi_bitfield algorithm
    
    Gil-free implementation of the compression algorithm
    
    :param frame: numpy.ndarray
    :returns bytes
    """
    cdef:
        Py_ssize_t shape, index
        uint8_t[::1] buffer    # Contains the compressed lines
        uint32_t[::1] cumsum
        int32_t [::1] delta     # buffer space used by compre_row
        uint32_t current
        uint16_t size
        uint64_t one=1
        
        
    shape = frame.shape[0]
    assert frame.shape[1] == shape, "Input shape is expected to be square !"
    
    buffer = numpy.empty(8*shape*shape, numpy.uint8) #Should be able to accomodate 4096² data 
    cumsum = numpy.empty(shape, numpy.uint32)
    delta = numpy.empty(shape-1, numpy.int32)

    with nogil:
        current = 0
        for index in range(shape):
            cumsum[index] = current
            size = _compress_row(frame[index], buffer[4+current:], delta)
            current += size
        buffer[0] = current & ((one<<8)-1)
        buffer[1] = (current & ((one<<16)-1))>>8
        buffer[2] = (current & ((one<<24)-1))>>16
        buffer[3] = (current & ((one<<32)-1))>>24

    if numpy.little_endian:
        return (numpy.asarray(buffer[:current+4]).tobytes()+
                numpy.asarray(cumsum).tobytes())
    else:
        return (numpy.asarray(buffer[:current+4]).tobytes()+
                numpy.asarray(cumsum).byteswap.tobytes())



# -*- coding: utf-8 -*-
#cython: embedsignature=True, language_level=3
## This is for optimisation
#cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False,
## This is for developping:
##cython: profile=True, warn.undeclared=True, warn.unused=True, warn.unused_result=False, warn.unused_arg=True
#
#    Project: Fable Input/Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) 2020-2024 European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Densification of sparse frame format
"""
__author__ = "Jérôme Kieffer"
__date__ = "19/03/2024"  
__contact__ = "Jerome.kieffer@esrf.fr"
__license__ = "MIT"

import time
import numpy
# from cython.parallel import prange
from libc.stdint cimport int8_t, uint8_t, \
                         uint16_t, int16_t,\
                         int32_t, uint32_t,\
                         int64_t, uint64_t
from libc.math cimport isfinite, log, sqrt, cos, M_PI                         
from libc.stdlib cimport RAND_MAX
cimport cython
               
ctypedef fused any_t:
    double
    float
    int8_t
    uint8_t
    uint16_t
    int16_t
    int32_t
    uint32_t
    int64_t
    uint64_t

#Few constants for 64-bit Mersenne Twisters
cdef:
    uint32_t NN = 312
    uint32_t MM = 156
    uint64_t MATRIX_A = 0xB5026F5AA96619E9ULL
    uint64_t UM = 0xFFFFFFFF80000000ULL # Most significant 33 bits
    uint64_t LM = 0x7FFFFFFFULL #Least significant 31 bits
    double NRM53 = 1.0 / ((1<<53)-1) #Normalisation for 53 bit integer
    double EPS64 = numpy.finfo(numpy.float64).eps
    double TWO_PI = 2.0 * M_PI
    

cdef class MT:
    """
    This class implements 64-bit Mersenne Twisters
    
    http://www.math.sci.hiroshima-u.ac.jp/m-mat/MT/VERSIONS/C-LANG/mt19937-64.c
    
    Inspired from:
    https://github.com/ananswam/cython_random
    with minor clean-ups
    
    Licence: MIT
    """
    cdef:
        uint64_t mt[312]
        uint32_t mti
        uint64_t mag01[2]
        bint has_spare
        double spare

    
    def __init__(self, seed):
        self.mti = NN + 1
        self._seed(<uint64_t> seed)
    
    cdef inline void _seed(self, uint64_t seed) noexcept nogil:
        self.mt[0] = seed
        for self.mti in range(1, NN):
            self.mt[self.mti] = (6364136223846793005ULL * (self.mt[self.mti-1] ^ (self.mt[self.mti-1] >> 62)) + self.mti)
        self.mag01[0] = 0ULL
        self.mag01[1] = MATRIX_A
        self.mti = NN
        self.has_spare = False
        
    cdef inline uint64_t genrand64(self) noexcept nogil:
        cdef: 
            uint32_t i
            uint64_t x
        if self.mti >= NN:
            for i in range(NN - MM):
                x = (self.mt[i]&UM) | (self.mt[i+1]&LM)
                self.mt[i] = self.mt[i+MM] ^ (x>>1) ^ self.mag01[int(x&1ULL)]

            for i in range(NN-MM, NN-1):
                x = (self.mt[i]&UM)|(self.mt[i+1]&LM)
                self.mt[i] = self.mt[i+(MM-NN)] ^ (x>>1) ^ self.mag01[int(x&1ULL)]

            x = (self.mt[NN-1]&UM)|(self.mt[0]&LM)
            self.mt[NN-1] = self.mt[MM-1] ^ (x>>1) ^ self.mag01[int(x&1ULL)]
            self.mti = 0

        x = self.mt[self.mti]
        self.mti += 1
        x ^= (x >> 29) & 0x5555555555555555ULL
        x ^= (x << 17) & 0x71D67FFFEDA60000ULL
        x ^= (x << 37) & 0xFFF7EEE000000000ULL
        x ^= (x >> 43);
        return x
    
    def rand(self):
        return self.genrand64()%(<uint64_t>RAND_MAX+1)
    
    @cython.cdivision(True)
    cdef inline double _uniform(self) noexcept nogil:
        return (self.genrand64() >> 11) * NRM53
    
    def uniform(self):
        "Return a random value between [0:1["
        return self._uniform()
    
    cdef inline double _normal_bm(self, double mu, double sigma) noexcept nogil:
        "Box-Muller implementation of the normal distribution"
        cdef:
            double u1=0.0, u2=0.0
    
        while (u1 == 0.0):
            u1 = self._uniform()
            u2 = self._uniform()

        return sigma * sqrt(-2.0 * log(u1)) * cos(TWO_PI * u2) + mu;

    cdef inline double _normal_m(self, double mu, double sigma) noexcept nogil:
        "Marsaglia implementation of the normal distribution, 2xfaster than Box-Muller"
        cdef: 
            double u1=0.0, u2=0.0, s=0.0
        if self.has_spare:
            self.has_spare = False
            return mu + self.spare * sigma 
        else:
            while (s>=1 or s==0.0):
                u1 = 2.0 * self._uniform() - 1.0
                u2 = 2.0 * self._uniform() - 1.0
                s = u1 * u1 + u2 * u2;
            s = sqrt(-2.0*log(s)/s)
            self.spare = u2 * s
            self.has_spare = True
        return mu + sigma * u1 * s;

    def normal(self, mu, sigma): 
        """
        Calculate the gaussian distribution using the Marsaglia algorithm

        Credits:
        https://en.wikipedia.org/wiki/Marsaglia_polar_method
        https://en.wikipedia.org/wiki/Box%E2%80%93Muller_transform

        :param mu: the center of the distribution
        :param sigma: the width of the distribution
        :return: random value
        """       
        return self._normal_m(mu, sigma)



def distribution_uniform_mtc(shape, seed=None):
    "Function to test uniform distribution"
    if seed is None:  
        try:
            seed = time.time_ns()
        except:
            seed = int(time.time()*1e9)
    cdef: 
        uint64_t size = numpy.prod(shape), idx
        double[::1] ary = numpy.empty(size)
        MT mt = MT(seed)
    with nogil:
        for idx in range(size):
            ary[idx] = mt._uniform()
    return numpy.asarray(ary).reshape(shape)      


def distribution_normal_mtc(mu, sigma, seed=None):
    "Function to test normal distribution"
    shape = mu.shape
    assert mu.shape == sigma.shape
    if seed is None:  
        try:
            seed = time.time_ns()
        except:
            seed = int(time.time()*1e9)
    cdef: 
        uint64_t size = numpy.prod(shape), idx
        double[::1] ary = numpy.empty(size)
        double[::1] cmu = numpy.ascontiguousarray(mu, dtype=numpy.float64).ravel()
        double[::1] csigma = numpy.ascontiguousarray(sigma, dtype=numpy.float64).ravel()
        MT mt = MT(seed)
    with nogil:
        for idx in range(size):
            ary[idx] = mt._normal_m(cmu[idx], csigma[idx])
    return numpy.asarray(ary).reshape(shape)        


def densify(cython.floating[:,::1] mask,
            cython.floating[::1] radius,
            uint32_t[::1] index,
            any_t[::1] intensity,
            any_t dummy,
            dtype,
            float[::1] background,
            float[::1] background_std=None,
            normalization=None,
            seed = None):
    """
    Densify a sparse representation to generate a normal frame 
    
    :param mask: 2D array with NaNs for mask and pixel radius for the valid pixels
    :param radius: 1D array with the radial distance
    :param background: 1D array with the background values at given distance from the center
    :param index: position of non-background pixels
    :param intensity: intensities of non background pixels (at index position)
    :param dummy: numerical value for masked-out pixels in dense image
    :param dtype: dtype of intensity.
    :param background_std: 1D array with the background std at given distance from the center --> activates the noisy mode.
    :param normalization: normalization array: renormalize all data with this factor (pixel-wise)
    :param seed: seed for the random number-generator, used only when regenerating noisy background
    :return: dense frame as 2D array
    """
    cdef:
        Py_ssize_t i, j, size, pos, size_over, width, height
        double value, fres, fpos, idelta, start, std
        bint integral, noisy, do_normalization=False, do_background=True
        any_t[:, ::1] dense
        float[:,::1] c_normalization
        MT mt
    if radius is None:
        do_background = False
    else:
        size = radius.shape[0]
        assert background.shape[0] == size
    size_over = index.shape[0]
    assert intensity.shape[0] == size_over
    integral = numpy.issubdtype(dtype, numpy.integer)
    height =mask.shape[0] 
    width = mask.shape[1]
    dense = numpy.zeros((height, width), dtype=dtype)
    if normalization is not None:
        do_normalization = True
        c_normalization = numpy.ascontiguousarray(normalization, dtype=numpy.float32)
    
    if background_std is None:
        noisy = False
    else:
        noisy=True
        if seed is None:
            try:
                seed = time.time_ns()
            except Exception:
                seed = int(time.time()*1e9)
        mt = MT(seed)
                
    with nogil:
        if do_background:
            start = radius[0]
            idelta = <double>(size - 1)/(radius[size-1] - start)  
            
            #Linear interpolation
            for i in range(height):
                for j in range(width):
                    fpos = (mask[i,j] - start)*idelta
                    if (fpos<0) or (fpos>=size) or (not isfinite(fpos)):
                        dense[i,j] = dummy 
                    else:
                        pos = <uint32_t> fpos
                        if pos+1 == size:
                            value = background[pos]
                            fres = 0.0
                        else:
                            fres = fpos - pos
                            value = (1.0 - fres)*background[pos] + fres*background[pos+1]
                        if noisy:
                            if pos+1 == size:
                                std = background_std[pos]
                                fres = 0.0
                            else:
                                std = (1.0 - fres)*background_std[pos] + fres*background_std[pos+1]
                            value = max(0.0, mt._normal_m(value, std))
                        if do_normalization:
                            value *= c_normalization[i, j]
                        if integral:
                            dense[i,j] =  <any_t>(value + 0.5) #this is rounding
                        else:
                            dense[i,j] =  <any_t>(value) 
        # Assignment of outliers
        for i in range(size_over):
            j = index[i]
            dense[j//width, j%width] = intensity[i]
    return numpy.asarray(dense) 